{*******************************************************************}
{    PColor 1.53                                                    }
{    Draw 2D data - flat, interp, isoline.                          }
{    Copyright (c) 1999-2001 Andrey Filatkin                        }
{    All Rights Reserved                                            }
{    anfilat@chat.ru ;   http://www.chat.ru/~anfilat                }
{    Last modified - 16 October 2001                                }
{*******************************************************************}
{    Thank to:                                                      }
{      Agustin (Qxo) Villena (avillena@dcc.uchile.cl)               }
{      and John Biddiscombe (J.Biddiscombe@rl.ac.uk)                }
{      for TDIBPaintBox and TDIBSurface (dibpbox.zip).              }
{*******************************************************************}
unit Pcolor;

interface

uses Windows, Classes, Graphics, Controls, Forms, SysUtils, DIBWork;

const
      { Greatest possible zoom. }
      MaxZoom = 1e6;

type
  PArray = ^TArray;
  { Dynamic array for the data. PColor draws these data.}
  TArray = array of array of double;
  { Possible palettes.
  First - 'cmLine' - the user sets first and last colors,
   others turn out by linear interpolation.
  Second - 'cmSpectr' - colors from blue up to red.
  Third - 'cmRandom' - random colors.
  Fourth - 'cmUser' - user defines colors.}
  TColorMaps = (cmLine, cmSpectr, cmRandom, cmUser);
  { Possible types of the image.
  'dtFlat' - each element is represented by a rectangular
   of the appropriate color.
  'dtInterp' - each point has color by turning out 
   interpolation of colors of four elements.
  'dtContour' - are draw isolines (section of the array
   by horizontal planes).
  'dtInterpContour' - isolines over interpolation of colors}
  TDrawTypes = (dtFlat, dtInterp, dtContour, dtInterpContour);
  { Possible modes of the image.
  'imnormal' - draw array,
  'imGradient' - draw gradient of array}
  TImageModes = (imNormal, imGradient);

  {Interface for work with Title.}
  TTitle = class(TPersistent)
  private
    FAlignment: TAlignment;
    FFont : TFont;
    FText : TCaption;
    FVisible : Boolean;
    FOnChange: TNotifyEvent;
    procedure SetAlignment(Value: TAlignment);
    procedure SetFont(Value: TFont);
    procedure SetText(const Value: TCaption);
    procedure SetVisible(Value: Boolean);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Value: TPersistent); override;
    procedure Changed;
  published
    {Controls the horizontal placement of the text.
    Set Alignment to specify how the text of the Title is aligned
    within PColor control.}
    property Alignment: TAlignment read FAlignment write SetAlignment default taCenter;
    {Controls the attributes of text title.}
    property Font : TFont read FFont write SetFont;
    {Specifies a text string that associated with the Title.}
    property Text : TCaption read FText write SetText;
    {Determines whether the Title appears on screen.
    Use the Visible property to control the visibility of the Title
    at runtime.}
    property Visible : Boolean read FVisible write SetVisible default True;
    {Occurs when the any property Title may have changed.}
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  {Interface for work with Axes.}
  TAxes = class(TPersistent)
  private
    FXMin, FXMax : Double;
    FYMin, FYMax : Double;
    FFont : TFont;
    FVisible : Boolean;
    FOnChange: TNotifyEvent;
    procedure SetFont(Value: TFont);
    procedure SetVisible(Value: Boolean);
  protected
  public
    constructor Create;
    destructor Destroy; override;
    procedure Assign(Value: TPersistent); override;
    procedure Changed;
    property XMax : Double read FXMax;
    property XMin : Double read FXMin;
    property YMax : Double read FYMax;
    property YMin : Double read FYMin;
  published
    property Font : TFont read FFont write SetFont;
    property Visible : Boolean read FVisible write SetVisible default True;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  {Interface for work with Coordinates.}
  TPColorCoordinates = class(TPersistent)
  private
    FAutomatic : Boolean;
    {Height of the Image (area of drawing DIB).}
    FImageHeight : Integer;
    {Left of the Image (area of drawing DIB).}
    FImageLeft : Integer;
    {Top of the Image (area of drawing DIB).}
    FImageTop : Integer;
    {Width of the Image (area of drawing DIB).}
    FImageWidth : Integer;
    FTitleHeight : Integer;
    FTitleLeft : Integer;
    FTitleTop : Integer;
    FTitleWidth : Integer;
    FColorBarHeight : Integer;
    FColorBarLeft : Integer;
    FColorBarTop : Integer;
    FColorBarWidth : Integer;
    FOnChange: TNotifyEvent;
    procedure SetAutomatic(Value: Boolean);
    procedure SetImageHeight(Value: Integer);
    procedure SetImageLeft(Value: Integer);
    procedure SetImageTop(Value: Integer);
    procedure SetImageWidth(Value: Integer);
    procedure SetTitleHeight(Value: Integer);
    procedure SetTitleLeft(Value: Integer);
    procedure SetTitleTop(Value: Integer);
    procedure SetTitleWidth(Value: Integer);
    procedure SetColorBarHeight(Value: Integer);
    procedure SetColorBarLeft(Value: Integer);
    procedure SetColorBarTop(Value: Integer);
    procedure SetColorBarWidth(Value: Integer);
  protected
  public
    constructor Create;
    procedure Assign(Value: TPersistent); override;
    procedure Changed;
  published
    property Automatic : Boolean read FAutomatic write SetAutomatic;
    property ImageHeight : Integer read FImageHeight write SetImageHeight;
    property ImageLeft : Integer read FImageLeft write SetImageLeft;
    property ImageTop : Integer read FImageTop write SetImageTop;
    property ImageWidth : Integer read FImageWidth write SetImageWidth;
    property TitleHeight : Integer read FTitleHeight write SetTitleHeight;
    property TitleLeft : Integer read FTitleLeft write SetTitleLeft;
    property TitleTop : Integer read FTitleTop write SetTitleTop;
    property TitleWidth : Integer read FTitleWidth write SetTitleWidth;
    property ColorBarHeight : Integer read FColorBarHeight write SetColorBarHeight;
    property ColorBarLeft : Integer read FColorBarLeft write SetColorBarLeft;
    property ColorBarTop : Integer read FColorBarTop write SetColorBarTop;
    property ColorBarWidth : Integer read FColorBarWidth write SetColorBarWidth;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  {TPColor - component made for visualization 2D arrays.
  The values of elements of the array define color used for their image.}
  TPColor = class(TCustomControl)
  private
    FAbout : String;
    FOnPaint: TNotifyEvent;
    FOnZoom : TNotifyEvent;
    FPCWidth: Integer;
    FPCHeight: Integer;
    FDrawFacet : boolean;
    FColorFacet : TColor;
    FColorFon : TColor;
    FColorMinValue : TColor;
    FColorMaxValue : TColor;
    FColorMap : TColorMaps;
    FTypeImage : TDrawTypes;
    FColorImageFon : TColor;
    FColorLower : TColor;
    FColorHigher : TColor;
    FContourLines : Integer;
    FColorContour : TColor;
    FDIB : TDIBWork;
    FLogPalette  : TLogPaletteDIB;
    FUserPalette : TLogPaletteDIB;
    NumberX : Integer;
    NumberY : Integer;
    NumberXm1 : Integer;
    NumberYm1 : Integer;
    FGraph : TArray;
    FMouseOn: Boolean;
    Origin : TPoint;
    MovePt : TPoint;
    FOnMouseDown : TMouseEvent;
    FOnMouseMove: TMouseMoveEvent;
    FOnMouseUp: TMouseEvent;
    FOldColor : Byte;
    Drawing : Boolean;
    ChangeGraph : Boolean;
    ChangePalette : Boolean;
    MaxMinValGraphValid : Boolean;
    GMaxMinValGraphValid : Boolean;
    IrregularData : Boolean;
    FXGrid : array of double;
    FYGrid : array of double;
    FXGrid2 : array of double;
    FYGrid2 : array of double;
    FDrawColorBar : Boolean;
    FColorBarKind : TScrollBarKind;
    ColorBarVis : Boolean;
    FNumbCol : Integer;

    FAxes : TAxes;
    FCoordinates : TPColorCoordinates;
    FTitle : TTitle;
    TitleVis : Boolean;

    FLogLeft : Double;
    FLogTop : Double;
    FLogRight : Double;
    FLogBottom : Double;
    FLogCentreX : Double;
    FLogCentreY : Double;
    FHorZoom : Double;
    FVerZoom : Double;
    FMaxMinValGraphAuto : Boolean;
    FMaxValGraph : Double;
    FMinValGraph : Double;
    FMaxValGraphX : Integer;
    FMaxValGraphY : Integer;
    FMinValGraphX : Integer;
    FMinValGraphY : Integer;
    MouseSensitivityX, MouseSensitivityY : Integer;
    FImageMode : TImageModes;
    FaHX, FaHY : Double;
    FGMaxMinValGraphAuto : Boolean;
    FGMaxValGraph : Double;
    FGMinValGraph : Double;
    FGMaxValGraphX : Integer;
    FGMaxValGraphY : Integer;
    FGMinValGraphX : Integer;
    FGMinValGraphY : Integer;

    procedure CreateImage;
    procedure CreatePaletta;
    procedure CreateFlat;
    procedure CreateInterp;
    procedure CreateContour;
    procedure SetAbout(AAbout : String);
    procedure SetColorFon(AColor : TColor);
    procedure SetColorImageFon(AColor : TColor);
    procedure SetColorMaxValue(AColor : TColor);
    procedure SetColorMinValue(AColor : TColor);
    procedure SetColorFacet(AColor : TColor);
    procedure SetColorLower(AColor : TColor);
    procedure SetColorHigher(AColor : TColor);
    procedure SetColorContour(AColor : TColor);
    procedure SetDrawFacet(ADrawFacet : Boolean);
    procedure SetDrawColorBar(ADraw : Boolean);
    procedure SetColorBarKind(AKind : TScrollBarKind);
    procedure SetColorMap(AColorMap : TColorMaps);
    procedure SetTypeImage(ATypeImage : TDrawTypes);
    procedure SetContourLines(AContourLines : Integer);
    procedure SetMouseOn(const Value: Boolean);
    procedure Sizes;
    procedure SetNumbCol(ANumbCol : Integer);
    procedure DrawCurColor(AColor : Integer);
    function  GetGraph(X, Y: Integer) : Double;
    procedure SetGraph(X, Y: Integer; Value: Double);
    function  GraphCI(Y, X: Integer) : Double;
    function  aHX(X: Integer) : Double;
    function  aHY(Y: Integer) : Double;
    function  GetUserPaletteEntry(Color: Byte): TPaletteEntry;
    procedure SetUserPaletteEntry(Color: Byte; Value: TPaletteEntry);

    procedure SetAxes(Value: TAxes);

    procedure SetTitle(Value: TTitle);
    procedure TitleChanged(Sender: TObject);
    procedure DoDrawText(var Rect: TRect; Flags: Longint; Canva: TCanvas);
    procedure SetCoordinates(Value: TPColorCoordinates);
    procedure CoordinatesChanged(Sender: TObject);

    function LogXToArr(Value : Double) : Double;
    function LogYToArr(Value : Double) : Double;
    function LogXToXGridFloor(Value : Double) : Integer;
    function LogXToXGridCeil(Value : Double) : Integer;
    function LogYToYGridFloor(Value : Double) : Integer;
    function LogYToYGridCeil(Value : Double) : Integer;
    procedure SetLogLeft(Value : Double);
    procedure SetLogTop(Value : Double);
    procedure SetLogRight(Value : Double);
    procedure SetLogBottom(Value : Double);
    procedure SetLogCentreX(Value : Double);
    procedure SetLogCentreY(Value : Double);
    procedure SetHorZoom(Value : Double);
    procedure SetVerZoom(Value : Double);
    procedure FindMaxMinValGraph;
    procedure SetMaxMinValGraphAuto(Value : Boolean);
    procedure SetMaxValGraph(Value : Double);
    function GetMaxValGraph : Double;
    procedure SetMinValGraph(Value : Double);
    function GetMinValGraph : Double;
    procedure SetImageMode(AImageMode : TImageModes);
    procedure FindGMaxMinValGraph;
    procedure SetGMaxMinValGraphAuto(Value : Boolean);
    procedure SetGMaxValGraph(Value : Double);
    function GetGMaxValGraph : Double;
    procedure SetGMinValGraph(Value : Double);
    function GetGMinValGraph : Double;
  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton;  Shift: TShiftState;
       X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton;  Shift: TShiftState;
       X, Y: Integer); override;
    procedure Zoom; virtual;
    procedure Draw(Canva : TCanvas);
    procedure DrawImage(Canva : TCanvas);
  public
    { Creates an instance of TPColor. Call Create to instantiate
    a TPColor object. Create allocates memory for the instance
    of TPColor.}
    constructor Create(AOwner: TComponent); override;
    { Destroys an instance of TPColor. Destroy frees the memory for
    the TPColor object.}
    destructor Destroy; override;
    procedure Assign(Value: TPersistent); override;
    property Canvas;
    {Is used care of a array, which should be drawn. The array before
    transfer should be filled with the data. AGraphAr - pointer on
    the array. XMin, XMax, YMin, YMax - logic coordinates attributed
    to a array.}
    procedure DrawData(AGraphAr : PArray; const XMin,XMax,YMin,YMax : Double);
    {Draw data with unequal grid spacing.}
    procedure DrawIrregularData(AGraphAr : PArray; const XGrid,YGrid : array of double);
    {Draw new data on the same grid.}
    function RefreshData(AGraphAr : PArray) : Boolean;
    {Erase Data. Size Graph - 0,0.}
    procedure SetDefaultData;
    {Draw only image}
    procedure RePaintImage;
    {True, if coordinate (APosImageX, APosImageY ) inside Image.}
    function IsPosInImage(APosImageX, APosImageY : Integer): Boolean;
    {Translates coordinates (in pixels) in logic coordinates.
    The logic coordinates for the whole array are set in DrawData
    or SetSizeGraph.}
    function PosImageToPosLog(APosImageX, APosImageY : Integer;
             var XAx, YAx : Double) : Boolean;
    {Translates logic coordinates in array coordinates.}
    function PosLogToPosArr(ALogX, ALogY : Double;
           var ArrX, ArrY : Double) : Boolean;
    {Translates array coordinates in logic coordinates.}
    function PosArrToPosLog(AArrX, AArrY : Double;
           var LogX, LogY : Double) : Boolean;
    {Returned interpolation value in definite position.}
    function ValueInPosLog(APosLogX, APosLogY : Double): Double;          
    {Logic coordinate appropriate to the bottom part of image.}
    property LogBottom : Double read FLogBottom write SetLogBottom;
    {Logic coordinate appropriate to the left part of image.}
    property LogLeft : Double read FLogLeft write SetLogLeft;
    {Logic coordinate appropriate to the right part of image.}
    property LogRight : Double read FLogRight write SetLogRight;
    {Logic coordinate appropriate to the top part of image.}
    property LogTop : Double read FLogTop write SetLogTop;
    {Logic coordinate appropriate to centre of image.}
    property LogCentreX : Double read FLogCentreX write SetLogCentreX;
    {Logic coordinate appropriate to centre of image.}
    property LogCentreY : Double read FLogCentreY write SetLogCentreY;
    {Shows in how many of time the image on a horizontal is increased.}
    property HorZoom : Double read FHorZoom write SetHorZoom;
    {Shows in how many of time the image on a vertical is increased.}
    property VerZoom : Double read FVerZoom write SetVerZoom;
    {The horizontal size of the array Graph.}
    property GraphSizeX : Integer read NumberX;
    {The vertical size of the array Graph.}
    property GraphSizeY : Integer read NumberY;
    {Establishes the size of the array, which is drawn PColor.
    For access to a array are used SetGraph and GetGraph.
    XMin, XMax, YMin, YMax - logic coordinates attributed
    to a array.}
    procedure SetSizeGraph(AGraphSizeX, AGraphSizeY : Integer;
                 const XMin,XMax,YMin,YMax : Double);
    {The array which is drawn PColor.
    The size of the array is established DrawData or SetSizeGraph.
    For access to the array are used SetGraph and GetGraph.}
    property Graph[X, Y: Integer]: Double read GetGraph write SetGraph;
    {The part of the array Graph allocated in logic coordinates
    ALeft, ATop, ARight, ABottom is drawn.}
    procedure ZoomArea(const ALeft,ATop,ARight,ABottom : Double);
    {The part of the array Graph, with centre in a point X, Y,
    with increase AHor on a horizontal and AVer on a vertical is drawn.}
    procedure ZoomAt(const X,Y : Double; const AHor,AVer : Double);
    {The part of the array Graph, with centre in a point X, Y,
    with current increase on a horizontal and on a vertical is drawn.}
    procedure ZoomAtPos(const X,Y : Double);
    {The part of the array Graph, with increase AHor on a horizontal
    and AVer on a vertical is drawn.}
    procedure ZoomAtTime(const AHor,AVer : Double);
    {Drawn all the array Graph. The increase is equal 1.}
    procedure ZoomOut;
    {SaveToFile writes the all graphic to a file, specified by FileName.}
    procedure SaveToFile(const FileName: String);
    {Creates a local copy of the all graphic in Clipboard format.}
    procedure SaveToClipBoard;
    {SaveImageToFile writes the Image to a file, specified by FileName.}
    procedure SaveImageToFile(const FileName: String);
    {Creates a local copy of the Image in Clipboard format.}
    procedure SaveImageToClipBoard;
    property UserPalette[Color: Byte]: TPaletteEntry read GetUserPaletteEntry write SetUserPaletteEntry;
  published
    property About : String read FAbout write SetAbout;
    property Align;
    property Anchors;
    property Axes : TAxes read FAxes write SetAxes;
    property Color;
    {ColorBarKind defines orientation ColorBar -
    horizontally or vertically.}
    property ColorBarKind : TScrollBarKind read FColorBarKind write SetColorBarKind default sbVertical;
    property ColorContour : TColor read FColorContour write SetColorContour default clBlack;
    property ColorFon : TColor read FColorFon write SetColorFon default clWhite;
    property ColorFacet : TColor read FColorFacet write SetColorFacet default clBlack;
    property ColorHigher : TColor read FColorHigher write SetColorHigher default clRed;
    property ColorImageFon : TColor read FColorImageFon write SetColorImageFon default clWhite;
    property ColorLower : TColor read FColorLower write SetColorLower default clBlue;
    property ColorMap : TColorMaps read FColorMap write SetColorMap default cmSpectr;
    property ColorMinValue : TColor read FColorMinValue write SetColorMinValue default clBlack;
    property ColorMaxValue : TColor read FColorMaxValue write SetColorMaxValue default clWhite;
    property Constraints;
    property ContourLines :Integer read FContourLines write SetContourLines default 10;
    property Coordinates : TPColorCoordinates read FCoordinates write SetCoordinates;
    property DragCursor;
    property DragKind;
    property DragMode;
    {ColorBar is color scale. ColorBar is drawn if DrawColorBar - true. }
    property DrawColorBar : Boolean read FDrawColorBar write SetDrawColorBar default True;
    property DrawFacet : Boolean read FDrawFacet write SetDrawFacet default True;
    property Enabled;
    property Font;
    property GMaxMinValGraphAuto : Boolean read FGMaxMinValGraphAuto write SetGMaxMinValGraphAuto default True;
    property GMaxValGraph : Double read GetGMaxValGraph write SetGMaxValGraph;
    property GMinValGraph : Double read GetGMinValGraph write SetGMinValGraph;
    property ImageMode : TImageModes read FImageMode write SetImageMode default imNormal;
    property MaxMinValGraphAuto : Boolean read FMaxMinValGraphAuto write SetMaxMinValGraphAuto default True;
    property MaxValGraph : Double read GetMaxValGraph write SetMaxValGraph;
    property MinValGraph : Double read GetMinValGraph write SetMinValGraph;
    property MouseOn : Boolean read FMouseOn write SetMouseOn default True;
    {Number of colors in the palette. }
    property NumbCol : Integer read FNumbCol write SetNumbCol default 64;
    property ParentColor;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property Title : TTitle read FTitle write SetTitle;
    property TypeImage : TDrawTypes read FTypeImage write SetTypeImage default dtFlat;
    property Visible;
    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnMouseDown : TMouseEvent read FOnMouseDown write FOnMouseDown;
    property OnMouseMove : TMouseMoveEvent read FOnMouseMove write FOnMouseMove;
    property OnMouseUp : TMouseEvent read FOnMouseUp write FOnMouseUp;
    property OnPaint : TNotifyEvent read FOnPaint write FOnPaint;
    property OnStartDock;
    property OnStartDrag;
    property OnZoom: TNotifyEvent read FOnZoom write FOnZoom;
  end;

implementation

uses ClipBrd, Math;

{ TTitle }

constructor TTitle.Create;
begin
  inherited Create;
  FFont := TFont.Create;
  FAlignment := taCenter;
  FVisible := True;
end;

destructor TTitle.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TTitle.Assign(Value: TPersistent);
begin
  if Value is TTitle then
  begin
    FAlignment := TTitle(Value).Alignment;
    FFont.Assign(TTitle(Value).Font);
    FText := TTitle(Value).Text;
    FVisible := TTitle(Value).Visible;
  end
  else inherited Assign(Value);
end;

procedure TTitle.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TTitle.SetAlignment(Value: TAlignment);
begin
  if Value <> FAlignment then
  begin
    FAlignment := Value;
    Changed;
  end;
end;

procedure TTitle.SetFont(Value: TFont);
begin
  if Value <> FFont then
  begin
    FFont.Assign(Value);
    Changed;
  end;
end;

procedure TTitle.SetText(const Value: TCaption);
begin
  if Value <> FText then
  begin
    FText := Value;
    Changed;
  end;
end;

procedure TTitle.SetVisible(Value: Boolean);
begin
  if Value <> FVisible then
  begin
    FVisible := Value;
    Changed;
  end;
end;

{ End TTitle }

{ TAxes }

constructor TAxes.Create;
begin
  inherited Create;
  FFont := TFont.Create;
  FVisible := True;
  FXMax := 100;
  FXMin := 0;
  FYMax := 100;
  FYMin := 0;
end;

destructor TAxes.Destroy;
begin
  FFont.Free;
  inherited Destroy;
end;

procedure TAxes.Assign(Value: TPersistent);
begin
  if Value is TAxes then
  begin
    FFont.Assign(TAxes(Value).Font);
    FVisible := TAxes(Value).Visible;
    FXMax := TAxes(Value).XMax;
    FXMin := TAxes(Value).XMin;
    FYMax := TAxes(Value).YMax;
    FYMin := TAxes(Value).YMin;
  end
  else inherited Assign(Value);
end;

procedure TAxes.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TAxes.SetFont(Value: TFont);
begin
  if Value <> FFont then
  begin
    FFont.Assign(Value);
    Changed;
  end;
end;

procedure TAxes.SetVisible(Value: Boolean);
begin
  if Value <> FVisible then
  begin
    FVisible := Value;
    Changed;
  end;
end;

{ End TAxes }

{ TPColorCoordinates }

constructor TPColorCoordinates.Create;
begin
  inherited Create;
  FAutomatic := True;
  FImageHeight := 126;
  FImageLeft := 3;
  FImageTop := 21;
  FImageWidth := 135;
  FTitleHeight := 11;
  FTitleLeft := 3;
  FTitleTop := 5;
  FTitleWidth := 146;
  FColorBarHeight := 126;
  FColorBarLeft := 141;
  FColorBarTop := 21;
  FColorBarWidth := 6;
end;

procedure TPColorCoordinates.Assign(Value: TPersistent);
begin
  if Value is TPColorCoordinates then
  begin
    Automatic := TPColorCoordinates(Value).Automatic;
    if Automatic = False then
    begin
      FImageHeight    := TPColorCoordinates(Value).ImageHeight;
      FImageLeft      := TPColorCoordinates(Value).ImageLeft;
      FImageTop       := TPColorCoordinates(Value).ImageTop;
      FImageWidth     := TPColorCoordinates(Value).ImageWidth;
      FTitleLeft      := TPColorCoordinates(Value).TitleLeft;
      FTitleTop       := TPColorCoordinates(Value).TitleTop;
      FTitleWidth     := TPColorCoordinates(Value).TitleWidth;
      FColorBarHeight := TPColorCoordinates(Value).ColorBarHeight;
      FColorBarLeft   := TPColorCoordinates(Value).ColorBarLeft;
      FColorBarTop    := TPColorCoordinates(Value).ColorBarTop;
      FColorBarWidth  := TPColorCoordinates(Value).ColorBarWidth;
      Changed;
    end
  end
  else inherited Assign(Value);
end;

procedure TPColorCoordinates.Changed;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TPColorCoordinates.SetAutomatic(Value: Boolean);
begin
  FAutomatic := Value;
  Changed;
end;

procedure TPColorCoordinates.SetImageHeight(Value: Integer);
begin
  if (not FAutomatic) and (FImageHeight <> Value) then
  begin
    if Value < 0 then  FImageHeight := 0
    else FImageHeight := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetImageLeft(Value: Integer);
begin
  if (not FAutomatic) and (FImageLeft <> Value) then
  begin
    FImageLeft := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetImageTop(Value: Integer);
begin
  if (not FAutomatic) and (FImageTop <> Value) then
  begin
    FImageTop := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetImageWidth(Value: Integer);
begin
  if (not FAutomatic) and (FImageWidth <> Value) then
  begin
    if Value < 0 then  FImageWidth := 0
    else FImageWidth := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetTitleHeight(Value: Integer);
begin
end;

procedure TPColorCoordinates.SetTitleLeft(Value: Integer);
begin
  if (not FAutomatic) and (FTitleLeft <> Value) then
  begin
    FTitleLeft := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetTitleTop(Value: Integer);
begin
  if (not FAutomatic) and (FTitleTop <> Value) then
  begin
    FTitleTop := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetTitleWidth(Value: Integer);
begin
  if (not FAutomatic) and (FTitleWidth <> Value) then
  begin
    FTitleWidth := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetColorBarHeight(Value: Integer);
begin
  if (not FAutomatic) and (FColorBarHeight <> Value) then
  begin
    FColorBarHeight := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetColorBarLeft(Value: Integer);
begin
  if (not FAutomatic) and (FColorBarLeft <> Value) then
  begin
    FColorBarLeft := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetColorBarTop(Value: Integer);
begin
  if (not FAutomatic) and (FColorBarTop <> Value) then
  begin
    FColorBarTop := Value;
    Changed;
  end;
end;

procedure TPColorCoordinates.SetColorBarWidth(Value: Integer);
begin
  if (not FAutomatic) and (FColorBarWidth <> Value) then
  begin
    FColorBarWidth := Value;
    Changed;
  end;
end;

{ End TPColorCoordinates }

{ TPcolor }

constructor TPcolor.Create(AOwner: TComponent);
var
   i,j : integer;
begin
  inherited Create(AOwner);
  ControlStyle := ControlStyle + [csReplicatable];
  Width := 150;
  Height := 150;
  FAbout := 'PColor 1.53. Filatkin Andrey, Saransk';

  FCoordinates := TPColorCoordinates.Create;
  FCoordinates.OnChange := CoordinatesChanged;

  FTitle := TTitle.Create;
  FTitle.OnChange := TitleChanged;
  FTitle.Font.OnChange := TitleChanged;
  FTitle.FText := 'Title PColor';

  FAxes := TAxes.Create;
  FLogLeft := FAxes.XMin;
  FLogTop := FAxes.YMax;
  FLogRight := FAxes.XMax;
  FLogBottom := FAxes.YMin;
  FLogCentreX := (FLogLeft + FLogRight) / 2;
  FLogCentreY := (FLogTop + FLogBottom) / 2;
  FHorZoom := 1;
  FVerZoom := 1;

  FColorFon := clWhite;
  NumberX := 0;
  NumberY := 0;
  FContourLines :=10;
  FColorImageFon := clBtnFace;
  FColorMap := cmSpectr;
  FColorMaxValue := clWhite;
  FColorLower := clBlue;
  FColorHigher := clRed;
  FDrawFacet := True;
  FDrawColorBar := True;
  FMaxMinValGraphAuto := True;
  FGMaxMinValGraphAuto := True;
  FColorBarKind := sbVertical;
  FNumbCol := 64;
  FMouseOn := True;
  Drawing := False;
  ChangeGraph := True;
  ChangePalette := True;
  MaxMinValGraphValid := False;
  GMaxMinValGraphValid := False;
  IrregularData := False;
  FOldColor := 0;
  MouseSensitivityX := GetSystemMetrics(SM_CXDOUBLECLK) div 2;
  MouseSensitivityY := GetSystemMetrics(SM_CYDOUBLECLK) div 2;
  for i := 1 to 252 do
  begin
    FUserPalette.palEntry[i].peRed := 0;
    FUserPalette.palEntry[i].peGreen := 0;
    FUserPalette.palEntry[i].peBlue := 0;
    FUserPalette.palEntry[i].peFlags := pc_Reserved;
  end;
  if csDesigning in ComponentState then
  begin
    NumberX := 10;
    NumberY := 10;
    NumberXm1 := NumberX - 1;
    NumberYm1 := NumberY - 1;
    FaHX := NumberXm1 / (2 * (FAxes.FXMax - FAxes.FXMin));
    FaHY := NumberYm1 / (2 * (FAxes.FYMax - FAxes.FYMin));
    SetLength(FGraph,NumberY,NumberX);
    for i := 0 to 9 do
      for j := 0 to 9 do
        FGraph[i,j] := i * i + j * j;
    MaxMinValGraphValid := False;
    GMaxMinValGraphValid := False;
  end;
  FDIB := TDIBWork.Create(FCoordinates.ImageWidth, FCoordinates.ImageHeight);
end;

destructor TPcolor.Destroy;
begin
  FDIB.Free;
  FTitle.Free;
  FAxes.Free;
  FCoordinates.Free;
  inherited Destroy;
end;

procedure TPcolor.Assign(Value: TPersistent);
begin
  if Value is TPcolor then
  begin
    FAxes.Assign(TPcolor(Value).Axes);
    FColorBarKind  := TPcolor(Value).ColorBarKind;
    FColorContour  := TPcolor(Value).ColorContour;
    FColorFon      := TPcolor(Value).ColorFon;
    FColorFacet    := TPcolor(Value).ColorFacet;
    FColorHigher   := TPcolor(Value).ColorHigher;
    FColorImageFon := TPcolor(Value).ColorImageFon;
    FColorLower    := TPcolor(Value).ColorLower;
    FColorMap      := TPcolor(Value).ColorMap;
    FColorMinValue := TPcolor(Value).ColorMinValue;
    FColorMaxValue := TPcolor(Value).ColorMaxValue;
    FContourLines  := TPcolor(Value).ContourLines;
    FDrawColorBar  := TPcolor(Value).DrawColorBar;
    FDrawFacet     := TPcolor(Value).DrawFacet;
    FCoordinates.Assign(TPcolor(Value).Coordinates);
    FGMaxMinValGraphAuto := TPcolor(Value).GMaxMinValGraphAuto;
    FGMaxValGraph  := TPcolor(Value).GMaxValGraph;
    FGMinValGraph  := TPcolor(Value).GMinValGraph;
    FImageMode     := TPcolor(Value).ImageMode;
    FMaxMinValGraphAuto := TPcolor(Value).MaxMinValGraphAuto;
    FMaxValGraph   := TPcolor(Value).MaxValGraph;
    FMinValGraph   := TPcolor(Value).MinValGraph;
    FMouseOn       := TPcolor(Value).MouseOn;
    FNumbCol       := TPcolor(Value).NumbCol;
    FTitle.Assign(TPcolor(Value).Title);
    FTypeImage     := TPcolor(Value).TypeImage;
  end
  else inherited Assign(Value);
end;

procedure TPcolor.Sizes;
var
  DC: HDC;
  Rect1: TRect;
  Temp : Integer;
const
  Margin = 3;
begin
  if not (csReading in ComponentState) then
  begin
    FPCWidth := Width;
    FPCHeight := Height;
    if FCoordinates.Automatic then
    begin
      FCoordinates.FTitleLeft := Margin;
      FCoordinates.FTitleWidth := Width - FCoordinates.FTitleLeft - Margin;
      FCoordinates.FTitleTop := Abs(Round(0.5*FTitle.Font.Height));
    end;
    Rect1 := Rect(FCoordinates.TitleLeft,FCoordinates.TitleTop,
     FCoordinates.TitleWidth+FCoordinates.TitleLeft,
     FCoordinates.TitleHeight+FCoordinates.TitleTop);
    DC := GetDC(0);
    Canvas.Handle := DC;
    DoDrawText(Rect1, (DT_EXPANDTABS or DT_CALCRECT or DT_WORDBREAK),Canvas);
    Canvas.Handle := 0;
    ReleaseDC(0, DC);
    FCoordinates.FTitleHeight := Rect1.Bottom - Rect1.Top;
    if FCoordinates.Automatic then
    begin
      if FTitle.Visible and (FTitle.Text<>'') then TitleVis := True
      else TitleVis := False;

      FCoordinates.FImageLeft := Margin;
      if TitleVis then FCoordinates.FImageTop := FCoordinates.TitleHeight
        + 2 * FCoordinates.TitleTop
      else FCoordinates.FImageTop := Margin;
      if FDrawColorBar then
      begin
        if ColorBarKind = sbVertical then
        begin
          FCoordinates.FImageHeight := Height - FCoordinates.FImageTop - Margin;
          FCoordinates.FColorBarTop := FCoordinates.FImageTop;
          FCoordinates.FColorBarHeight := FCoordinates.FImageHeight;
          FCoordinates.FColorBarWidth := Trunc(0.1 * FCoordinates.FColorBarHeight);
          if FCoordinates.FColorBarWidth > 30 then FCoordinates.FColorBarWidth := 30;
          FCoordinates.FImageWidth := Width - FCoordinates.FImageLeft
            - FCoordinates.FColorBarWidth - 2 * Margin;
          Temp := Trunc(0.05 * FCoordinates.FImageWidth);
          if Temp < FCoordinates.FColorBarWidth then
          begin
            FCoordinates.FColorBarWidth := Temp;
            FCoordinates.FImageWidth := Width - FCoordinates.FImageLeft
              - FCoordinates.FColorBarWidth - 2 * Margin;
          end;
          FCoordinates.FColorBarLeft := FCoordinates.FImageLeft
            + FCoordinates.FImageWidth + Margin;
        end
        else
        begin
          FCoordinates.FImageWidth := Width - FCoordinates.FImageLeft - Margin;
          FCoordinates.FColorBarLeft := FCoordinates.FImageLeft;
          FCoordinates.FColorBarWidth := FCoordinates.FImageWidth;
          FCoordinates.FColorBarHeight := Trunc(0.1 * FCoordinates.FColorBarWidth);
          if FCoordinates.FColorBarHeight > 30 then FCoordinates.FColorBarHeight := 30;
          FCoordinates.FImageHeight := Height - FCoordinates.FImageTop
            - FCoordinates.FColorBarHeight - 2 * Margin;
          Temp := Trunc(0.05 * FCoordinates.FImageHeight);
          if Temp < FCoordinates.FColorBarHeight then
          begin
            FCoordinates.FColorBarHeight := Temp;
            FCoordinates.FImageHeight := Height - FCoordinates.FImageTop
              - FCoordinates.FColorBarHeight - 2 * Margin;
          end;
          FCoordinates.FColorBarTop := FCoordinates.FImageTop
            + FCoordinates.FImageHeight + Margin;
        end;
      end
      else
      begin
        FCoordinates.FImageWidth := Width - FCoordinates.FImageLeft - Margin;
        FCoordinates.FImageHeight := Height - FCoordinates.FImageTop - Margin;
      end;
      if (FCoordinates.FImageWidth < 0) or (FCoordinates.FImageHeight < 0) then
      begin
        FCoordinates.FImageWidth := 0;
        FCoordinates.FImageHeight := 0;
      end;
      if Height < (FCoordinates.TitleHeight + FCoordinates.TitleTop)
      then TitleVis := False;
      if FDrawColorBar and (FCoordinates.FImageWidth <> 0)
         and (FCoordinates.FImageHeight <> 0) then ColorBarVis := True
      else ColorBarVis := False;
    end
    else
    begin
      TitleVis := FTitle.Visible;
      ColorBarVis := FDrawColorBar;
    end;
    if (FCoordinates.FImageWidth <> FDIB.Width)
      or (FCoordinates.FImageHeight <> FDIB.Height) then
    begin
      FDIB.Resize(FCoordinates.ImageWidth, FCoordinates.ImageHeight);
      CreatePaletta;
      CreateImage;
    end;
  end;
end;

procedure TPcolor.Paint;
begin
  Canvas.Font := Font;
  Canvas.Brush.Color := Color;
  if (FPCWidth<>Width) or (FPCHeight<>Height) then Sizes;
  if ChangePalette then CreatePaletta;
  if ChangeGraph then CreateImage;
  Draw(Canvas);
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

procedure TPcolor.Draw(Canva : TCanvas);
const
  Alignments: array[TAlignment] of Word = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  Rect1 : TRect;
  DrawStyle: Longint;
  i : Integer;
  XStep, YStep : double;
begin
  with Canva do
  begin
    Brush.Style := bsSolid;
    Brush.Color := FColorFon;
    FillRect(Rect(0, 0, Width, Height));
    DrawImage(Canva);
    if ColorBarVis then
      if ColorBarKind = sbVertical then
      begin
        Brush.Style := bsSolid;
        YStep := FCoordinates.ColorBarHeight / (NumbCol + 2);
        Brush.Color := ColorLower;
        FillRect(Rect(FCoordinates.ColorBarLeft,
           FCoordinates.ColorBarTop + Trunc((NumbCol + 1) * YStep),
           FCoordinates.ColorBarLeft + FCoordinates.ColorBarWidth,
           FCoordinates.ColorBarTop + FCoordinates.ColorBarHeight));
        for i := 1 to NumbCol do
        begin
          Brush.Color := FLogPalette.palEntry[i].peRed
             or (FLogPalette.palEntry[i].peGreen shl 8)
             or (FLogPalette.palEntry[i].peBlue shl 16);
          FillRect(Rect(FCoordinates.ColorBarLeft,
             FCoordinates.ColorBarTop + Trunc((NumbCol + 1 - i) * YStep),
             FCoordinates.ColorBarLeft + FCoordinates.ColorBarWidth,
             FCoordinates.ColorBarTop + Trunc((NumbCol + 2 - i) * YStep)));
        end;
        Brush.Color := ColorHigher;
        FillRect(Rect(FCoordinates.ColorBarLeft,
           FCoordinates.ColorBarTop,
           FCoordinates.ColorBarLeft + FCoordinates.ColorBarWidth,
           FCoordinates.ColorBarTop + Trunc(YStep)));
      end
      else
      begin
        Brush.Style := bsSolid;
        XStep := FCoordinates.ColorBarWidth / (NumbCol + 2);
        Brush.Color := ColorLower;
        FillRect(Rect(FCoordinates.ColorBarLeft,
           FCoordinates.ColorBarTop,
           FCoordinates.ColorBarLeft + Trunc(XStep),
           FCoordinates.ColorBarTop + FCoordinates.ColorBarHeight));
        for i := 1 to NumbCol do
        begin
          Brush.Color:=FLogPalette.palEntry[i].peRed
             or (FLogPalette.palEntry[i].peGreen shl 8)
             or (FLogPalette.palEntry[i].peBlue shl 16);
          FillRect(Rect(FCoordinates.ColorBarLeft + Trunc(i * XStep),
             FCoordinates.ColorBarTop,
             FCoordinates.ColorBarLeft + Trunc((i + 1) * XStep),
             FCoordinates.ColorBarTop + FCoordinates.ColorBarHeight));
        end;
        Brush.Color := ColorHigher;
        FillRect(Rect(FCoordinates.ColorBarLeft + Trunc((NumbCol + 1) * XStep),
           FCoordinates.ColorBarTop,
           FCoordinates.ColorBarLeft + FCoordinates.ColorBarWidth,
           FCoordinates.ColorBarTop + FCoordinates.ColorBarHeight));
      end;
    if (FMouseOn) and (Canvas.Handle=Canva.Handle) then
      DrawCurColor(FOldColor);
    if TitleVis then
    begin
      Brush.Style := bsClear;
      Rect1 := Rect(FCoordinates.TitleLeft,
       FCoordinates.TitleTop,
       FCoordinates.TitleWidth + FCoordinates.TitleLeft,
       FCoordinates.TitleHeight + FCoordinates.TitleTop);
      DrawStyle := DT_EXPANDTABS or DT_WORDBREAK or Alignments[FTitle.Alignment];
      DoDrawText(Rect1, DrawStyle, Canva);
    end;
  end;
end;

procedure TPcolor.RePaintImage;
begin
  if ChangePalette then CreatePaletta;
  if ChangeGraph then CreateImage;
  DrawImage(Canvas);
  if Assigned(FOnPaint) then FOnPaint(Self);
end;

procedure TPcolor.DrawImage(Canva : TCanvas);
var
  i, temp : Integer;
  XStep, YStep, XL, XR, YT, YB : Double;
begin
  with Canva do
  begin
    if (NumberX <> 0) and (NumberY <> 0) then
    begin
      FDIB.Paint(Handle,FCoordinates.ImageLeft,FCoordinates.ImageTop);
      if FDrawFacet then
      begin
        Pen.Width := 1;
        Pen.Style := psSolid;
        Pen.Mode  := pmCopy;
        Pen.Color := FColorFacet;
        if IrregularData then
        begin
          for i := LogXToXGridCeil(LogLeft) to LogXToXGridFloor(LogRight) do
            begin
              temp := Round((FCoordinates.ImageWidth - 1) *
               (FXGrid[i + 1] - Logleft) / (LogRight - LogLeft));
              MoveTo(FCoordinates.ImageLeft+temp,FCoordinates.ImageTop);
              LineTo(FCoordinates.ImageLeft+temp,FCoordinates.ImageTop+FCoordinates.ImageHeight);
            end;
          for i := LogYToYGridCeil(LogBottom) to LogYToYGridFloor(LogTop) do
            begin
              temp := Round((FCoordinates.ImageHeight - 1) *
               (LogTop - FYGrid[i + 1]) / (LogTop - LogBottom));
              MoveTo(FCoordinates.ImageLeft,FCoordinates.ImageTop+temp);
              LineTo(FCoordinates.ImageLeft+FCoordinates.ImageWidth,FCoordinates.ImageTop+temp);
            end;
        end
        else
        begin
          XStep := HorZoom * (FCoordinates.ImageWidth-1) / NumberXm1;
          XL := LogXToArr(LogLeft);
          XR := LogXToArr(LogRight);
          for i := (maxint-Trunc(maxint-XL)) to Trunc(XR) do
            begin
              temp := Round((i-XL)*XStep);
              MoveTo(FCoordinates.ImageLeft+temp,FCoordinates.ImageTop);
              LineTo(FCoordinates.ImageLeft+temp,FCoordinates.ImageTop+FCoordinates.ImageHeight);
            end;
          YStep := VerZoom * (FCoordinates.ImageHeight-1) / NumberYm1;
          YT := LogYToArr(LogTop);
          YB := LogYToArr(LogBottom);
          for i := (maxint-Trunc(maxint-YB)) to Trunc(YT) do
            begin
              temp := Round((YT-i)*YStep);
              MoveTo(FCoordinates.ImageLeft,FCoordinates.ImageTop+temp);
              LineTo(FCoordinates.ImageLeft+FCoordinates.ImageWidth,FCoordinates.ImageTop+temp);
            end;
        end;
      end;
    end
    else
    begin
      Brush.Color := FColorImageFon;
      FillRect(Rect(FCoordinates.ImageLeft, FCoordinates.ImageTop,
       FCoordinates.ImageWidth+FCoordinates.ImageLeft,
       FCoordinates.ImageHeight+FCoordinates.ImageTop));
    end;
  end;
end;

procedure TPcolor.FindMaxMinValGraph;
var
  TempVal : Double;
  i,j : Integer;
begin
  if MaxMinValGraphAuto and (NumberX <> 0) and (NumberY <> 0) then
  begin
    FMaxValGraph := FGraph[0, 0];
    FMinValGraph := FGraph[0, 0];
    FMaxValGraphX := 0;
    FMaxValGraphY := 0;
    FMinValGraphX := 0;
    FMinValGraphY := 0;
    for i := 0 to NumberYm1 do
      for j := 0 to NumberXm1 do
      begin
        TempVal := FGraph[i,j];
        if (FMaxValGraph < TempVal) then
        begin
          FMaxValGraph := TempVal;
          FMaxValGraphX := j;
          FMaxValGraphY := i;
        end;
        if (FMinValGraph > TempVal) then
        begin
          FMinValGraph := TempVal;
          FMinValGraphX := j;
          FMinValGraphY := i;
        end;
      end;
    MaxMinValGraphValid := True;
  end;
end;

procedure TPcolor.SetMaxMinValGraphAuto(Value : Boolean);
var
  tMaxVal, tMinVal : Double;
begin
  if MaxMinValGraphAuto <> Value then
  begin
    FMaxMinValGraphAuto := Value;
    if MaxMinValGraphAuto then
    begin
      if ImageMode=imNormal then
      begin
        tMaxVal := FMaxValGraph;
        tMinVal := FMinValGraph;
        FindMaxMinValGraph;
        if (tMaxVal <> MaxValGraph) or (tMinVal <> MinValGraph) then
        begin
          ChangeGraph := True;
          RePaintImage;
        end;
      end
      else MaxMinValGraphValid := False;
    end;
  end;
end;

procedure TPcolor.SetMaxValGraph(Value : Double);
begin
  if (not MaxMinValGraphAuto) and (FMaxValGraph <> Value) then
  begin
    FMaxValGraph := Value;
    if FMaxValGraph < FMinValGraph then
      FMinValGraph := FMaxValGraph;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

function TPcolor.GetMaxValGraph : Double;
begin
  if MaxMinValGraphAuto and (not MaxMinValGraphValid) then FindMaxMinValGraph;
  Result := FMaxValGraph;
end;

procedure TPcolor.SetMinValGraph(Value : Double);
begin
  if (not MaxMinValGraphAuto) and (FMinValGraph <> Value) then
  begin
    FMinValGraph := Value;
    if FMinValGraph > FMaxValGraph then
      FMaxValGraph := FMinValGraph;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

function TPcolor.GetMinValGraph : Double;
begin
  if MaxMinValGraphAuto and (not MaxMinValGraphValid) then FindMaxMinValGraph;
  Result := FMinValGraph;
end;

procedure TPcolor.FindGMaxMinValGraph;
var
  TempVal, pX, pY : Double;
  i,j : Integer;
begin
  if GMaxMinValGraphAuto and (NumberX <> 0) and (NumberY <> 0) then
  begin
    if (NumberX > 3) and (NumberY > 3) then
    begin
      TempVal := Hypot((FGraph[1, 2] - FGraph[1, 0]) * aHX(1),
                       (FGraph[2, 1] - FGraph[0, 1]) * aHY(1));
      FGMaxValGraph := TempVal;
      FGMinValGraph := TempVal;
      FGMaxValGraphX := 1;
      FGMaxValGraphY := 1;
      FGMinValGraphX := 1;
      FGMinValGraphY := 1;
      for i := 1 to NumberY - 2 do
        for j := 1 to NumberX - 2 do
        begin
          TempVal := Hypot((FGraph[i, j + 1] - FGraph[i, j - 1]) * aHX(j),
                           (FGraph[i + 1, j] - FGraph[i - 1, j]) * aHY(i));
          if (FGMaxValGraph < TempVal) then
          begin
            FGMaxValGraph := TempVal;
            FGMaxValGraphX := j;
            FGMaxValGraphY := i;
          end;
          if (FGMinValGraph > TempVal) then
          begin
            FGMinValGraph := TempVal;
            FGMinValGraphX := j;
            FGMinValGraphY := i;
          end;
        end;
    end
    else
    begin
      TempVal := Hypot((FGraph[0, 1] - FGraph[0, 0]) * aHX(0),
                       (FGraph[1, 0] - FGraph[0, 0]) * aHY(0));
      FGMaxValGraph := TempVal;
      FGMinValGraph := TempVal;
      FGMaxValGraphX := 0;
      FGMaxValGraphY := 0;
      FGMinValGraphX := 0;
      FGMinValGraphY := 0;
      for i := 0 to NumberYm1 do
        for j := 0 to NumberXm1 do
        begin
          if j = 0 then
            pX := (FGraph[i, 1] - FGraph[i, 0]) * aHX(j)
          else if j = NumberXm1 then
            pX := (FGraph[i, NumberXm1] - FGraph[i, NumberXm1 - 1]) * aHX(j)
          else
            pX := (FGraph[i, j + 1] - FGraph[i, j - 1]) * aHX(j);
          if i = 0 then
            pY := (FGraph[1, j] - FGraph[0, j]) * aHY(i)
          else if i = NumberYm1 then
            pY := (FGraph[NumberYm1, j] - FGraph[NumberYm1 - 1, j]) * aHY(i)
          else
            pY := (FGraph[i + 1, j] - FGraph[i - 1, j]) * aHY(i);
          TempVal := Hypot(pX, pY);
          if (FGMaxValGraph < TempVal) then
          begin
            FGMaxValGraph := TempVal;
            FGMaxValGraphX := j;
            FGMaxValGraphY := i;
          end;
          if (FGMinValGraph > TempVal) then
          begin
            FGMinValGraph := TempVal;
            FGMinValGraphX := j;
            FGMinValGraphY := i;
          end;
        end;
    end;
    GMaxMinValGraphValid := True;
  end;
end;

procedure TPcolor.SetGMaxMinValGraphAuto(Value : Boolean);
var
  tMaxVal, tMinVal : Double;
begin
  if GMaxMinValGraphAuto <> Value then
  begin
    FGMaxMinValGraphAuto := Value;
    if GMaxMinValGraphAuto then
    begin
      if ImageMode=imGradient then
      begin
        tMaxVal := FGMaxValGraph;
        tMinVal := FGMinValGraph;
        FindGMaxMinValGraph;
        if (tMaxVal <> GMaxValGraph) or (tMinVal <> GMinValGraph) then
        begin
          ChangeGraph := True;
          RePaintImage;
        end;
      end
      else GMaxMinValGraphValid := False;
    end;
  end;
end;

procedure TPcolor.SetGMaxValGraph(Value : Double);
begin
  if (not GMaxMinValGraphAuto) and (FGMaxValGraph <> Value) then
  begin
    FGMaxValGraph := Value;
    if FGMaxValGraph < FGMinValGraph then
      FGMinValGraph := FGMaxValGraph;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

function TPcolor.GetGMaxValGraph : Double;
begin
  if GMaxMinValGraphAuto and (not GMaxMinValGraphValid) then FindGMaxMinValGraph;
  Result := FGMaxValGraph;
end;

procedure TPcolor.SetGMinValGraph(Value : Double);
begin
  if (not GMaxMinValGraphAuto) and (FGMinValGraph <> Value) then
  begin
    FGMinValGraph := Value;
    if FGMinValGraph > FGMaxValGraph then
      FGMaxValGraph := FGMinValGraph;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

function TPcolor.GetGMinValGraph : Double;
begin
  if GMaxMinValGraphAuto and (not GMaxMinValGraphValid) then FindGMaxMinValGraph;
  Result := FGMinValGraph;
end;

procedure TPcolor.DrawData(AGraphar : PArray; const XMin,XMax,YMin,YMax : Double);
var
   i,j,NX,NY : integer;
begin
  NX := High(AGraphar^)+1;
  NY := High(AGraphar^[0])+1;

  if (NX = 0) or (NY = 0) then
    SetDefaultData
  else
  begin
    ChangeGraph := True;
    IrregularData := False;
    if NX = 1 then NumberX := 2 else NumberX := NX;
    if NY = 1 then NumberY := 2 else NumberY := NY;
    NumberXm1 := NumberX - 1;
    NumberYm1 := NumberY - 1;
    FGraph := nil;
    FXGrid := nil;
    FYGrid := nil;
    FXGrid2 := nil;
    FYGrid2 := nil;
    SetLength(FGraph,NumberY,NumberX);
    for i := 0 to NY - 1 do
      for j := 0 to NX - 1 do FGraph[i,j] := AGraphar^[j,i];
    if NX = 1 then
      for i := 0 to NumberYm1 do FGraph[i,1] := FGraph[i,0];
    if NY = 1 then
      for i := 0 to NumberXm1 do FGraph[1,i] := FGraph[0,i];
    MaxMinValGraphValid := False;
    GMaxMinValGraphValid := False;

    if (XMin<>FAxes.XMin) or (XMax<>FAxes.XMax)
        or (YMin<>FAxes.YMin) or (YMax<>FAxes.YMax) then
    begin
      if XMin < XMax then
        begin  FAxes.FXMin := XMin;   FAxes.FXMax := XMax;  end
      else
        begin  FAxes.FXMin := XMax;   FAxes.FXMax := XMin;  end;
      if XMin = XMax then
        if XMin > 0 then FAxes.FXMax := XMin * 1.01
        else if XMin < 0 then FAxes.FXMax := XMin * 0.99
        else FAxes.FXMax := 0.01;
      if YMin < YMax then
        begin  FAxes.FYMin := YMin;   FAxes.FYMax := YMax;  end
      else
        begin  FAxes.FYMin := YMax;   FAxes.FYMax := YMin;  end;
      if YMin = YMax then
        if YMin > 0 then FAxes.FYMax := YMin * 1.01
        else if YMin < 0 then FAxes.FYMax := YMin * 0.99
        else FAxes.FYMax := 0.01;
      FaHX := NumberXm1 / (2 * (FAxes.FXMax - FAxes.FXMin));
      FaHY := NumberYm1 / (2 * (FAxes.FYMax - FAxes.FYMin));
      ZoomOut;
    end
    else
    begin
      FaHX := NumberXm1 / (2 * (FAxes.FXMax - FAxes.FXMin));
      FaHY := NumberYm1 / (2 * (FAxes.FYMax - FAxes.FYMin));
      RePaintImage;
    end;
  end;
end;

procedure TPcolor.DrawIrregularData(AGraphAr : PArray; const XGrid,YGrid : array of double);
var
   i,j,NX,NY,XGridN,YGridN : integer;
   XMin,XMax,YMin,YMax : double;
begin
  NX := High(AGraphar^)+1;
  NY := High(AGraphar^[0])+1;
  XGridN := High(XGrid)+1;
  YGridN := High(YGrid)+1;

  if (NX = 0) or (NY = 0) or (NX <> XGridN) or (NY <> YGridN) then
    SetDefaultData
  else
  begin
    ChangeGraph := True;
    IrregularData := True;
    XMin := XGrid[0];
    XMax := XGrid[XGridN - 1];
    YMin := YGrid[0];
    YMax := YGrid[YGridN - 1];
    if NX = 1 then NumberX := 2 else NumberX := NX;
    if NY = 1 then NumberY := 2 else NumberY := NY;
    NumberXm1 := NumberX - 1;
    NumberYm1 := NumberY - 1;
    FGraph := nil;
    SetLength(FGraph, NumberY, NumberX);
    SetLength(FXGrid, NumberX + 2);
    SetLength(FYGrid, NumberY + 2);
    SetLength(FXGrid2, NumberX);
    SetLength(FYGrid2, NumberY);
    for i := 1 to NX do FXGrid[i] := XGrid[i - 1];
    for i := 1 to NY do FYGrid[i] := YGrid[i - 1];
    for i := 0 to NY - 1 do
      for j := 0 to NX - 1 do FGraph[i,j] := AGraphar^[j,i];
    if NX = 1 then
    begin
      FXGrid[2] := FXGrid[1];
      for i := 0 to NumberYm1 do FGraph[i,1] := FGraph[i,0];
    end;
    if NY = 1 then
    begin
      FYGrid[2] := FYGrid[1];
      for i := 0 to NumberXm1 do FGraph[1,i] := FGraph[0,i];
    end;
    FXGrid[0] := 2 * FXGrid[1] - FXGrid[2];
    FXGrid[NumberX + 1] := 2 * FXGrid[NumberX] - FXGrid[NumberX - 1];
    FYGrid[0] := 2 * FYGrid[1] - FYGrid[2];
    FYGrid[NumberY + 1] := 2 * FYGrid[NumberY] - FYGrid[NumberY - 1];
    for i := 0 to NumberXm1 do FXGrid2[i] := 1 / (FXGrid[i + 2] - FXGrid[i]);
    for i := 0 to NumberYm1 do FYGrid2[i] := 1 / (FYGrid[i + 2] - FYGrid[i]);
    MaxMinValGraphValid := False;
    GMaxMinValGraphValid := False;

    if (XMin<>FAxes.XMin) or (XMax<>FAxes.XMax)
        or (YMin<>FAxes.YMin) or (YMax<>FAxes.YMax) then
    begin
      if XMin < XMax then
        begin  FAxes.FXMin := XMin;   FAxes.FXMax := XMax;  end
      else
        begin  FAxes.FXMin := XMax;   FAxes.FXMax := XMin;  end;
      if XMin = XMax then
        if XMin > 0 then FAxes.FXMax := XMin * 1.01
        else if XMin < 0 then FAxes.FXMax := XMin * 0.99
        else FAxes.FXMax := 0.01;
      if YMin < YMax then
        begin  FAxes.FYMin := YMin;   FAxes.FYMax := YMax;  end
      else
        begin  FAxes.FYMin := YMax;   FAxes.FYMax := YMin;  end;
      if YMin = YMax then
        if YMin > 0 then FAxes.FYMax := YMin * 1.01
        else if YMin < 0 then FAxes.FYMax := YMin * 0.99
        else FAxes.FYMax := 0.01;
      ZoomOut;
    end
    else
      RePaintImage;
  end;
end;

function TPcolor.RefreshData(AGraphAr : PArray) : Boolean;
var
   i,j,NX,NY,NX2,NY2 : integer;
begin
  NX := High(AGraphar^)+1;
  NY := High(AGraphar^[0])+1;
  if NX = 1 then NX2 := 2 else NX2 := NX;
  if NY = 1 then NY2 := 2 else NY2 := NY;

  if (NX2 <> NumberX) or (NY2 <> NumberY) or (NumberX = 0) or (NumberY = 0) then
    Result := False
  else
  begin
    ChangeGraph := True;
    for i := 0 to NY - 1 do
      for j := 0 to NX - 1 do FGraph[i,j] := AGraphar^[j,i];
    if NX = 1 then
      for i := 0 to NumberYm1 do FGraph[i,1] := FGraph[i,0];
    if NY = 1 then
      for i := 0 to NumberXm1 do FGraph[1,i] := FGraph[0,i];
    MaxMinValGraphValid := False;
    GMaxMinValGraphValid := False;

    RePaintImage;
    Result := True
  end;
end;

function TPcolor.GetGraph(X, Y: Integer): Double;
begin
  GetGraph := FGraph[Y, X];
end;

procedure TPcolor.SetGraph(X, Y: Integer; Value: Double);
var
  TempVal, pX, pY : Double;
begin
  FGraph[Y, X] := Value;
  ChangeGraph := True;
  if MaxMinValGraphAuto then
  begin
    if ((X=FMaxValGraphX) and (Y=FMaxValGraphY)) or
       ((X=FMinValGraphX) and (Y=FMinValGraphY)) then
      MaxMinValGraphValid := False
    else if MaxMinValGraphValid then
    begin
      if (FMaxValGraph < Value) then
      begin
        FMaxValGraph := Value;
        FMaxValGraphX := X;
        FMaxValGraphY := Y;
      end;
      if (FMinValGraph > Value) then
      begin
        FMinValGraph := Value;
        FMinValGraphX := X;
        FMinValGraphY := Y;
      end;
    end;
  end;
  if GMaxMinValGraphAuto then
  begin
    if ((X=FGMaxValGraphX) and (Y=FGMaxValGraphY)) or
       (((X-1)=FGMaxValGraphX) and (Y=FGMaxValGraphY)) or
       (((X+1)=FGMaxValGraphX) and (Y=FGMaxValGraphY)) or
       ((X=FGMaxValGraphX) and ((Y-1)=FGMaxValGraphY)) or
       ((X=FGMaxValGraphX) and ((Y+1)=FGMaxValGraphY)) or
       ((X=FGMinValGraphX) and (Y=FGMinValGraphY)) or
       (((X-1)=FGMinValGraphX) and (Y=FGMinValGraphY)) or
       (((X+1)=FGMinValGraphX) and (Y=FGMinValGraphY)) or
       ((X=FGMinValGraphX) and ((Y-1)=FGMinValGraphY)) or
       ((X=FGMinValGraphX) and ((Y+1)=FGMinValGraphY)) then
      GMaxMinValGraphValid := False
    else if GMaxMinValGraphValid then
    begin
      if X = 0 then
        pX := (FGraph[Y, 1] - FGraph[Y, 0]) * aHX(X)
      else if X = NumberXm1 then
        pX := (FGraph[Y, NumberXm1] - FGraph[Y, NumberXm1 - 1]) * aHX(X)
      else
        pX := (FGraph[Y, X + 1] - FGraph[Y, X - 1]) * aHX(X);
      if Y = 0 then
        pY := (FGraph[1, X] - FGraph[0, X]) * aHY(Y)
      else if Y = NumberYm1 then
        pY := (FGraph[NumberYm1, X] - FGraph[NumberYm1 - 1, X]) * aHY(Y)
      else
        pY := (FGraph[Y + 1, X] - FGraph[Y - 1, X]) * aHY(Y);
      TempVal := Hypot(pX, pY);
      if (FGMaxValGraph < TempVal) then
      begin
        FGMaxValGraph := TempVal;
        FGMaxValGraphX := X;
        FGMaxValGraphY := Y;
      end;
      if (FGMinValGraph > TempVal) then
      begin
        FGMinValGraph := TempVal;
        FGMinValGraphX := X;
        FGMinValGraphY := Y;
      end;
    end;
  end;
end;

function TPcolor.GraphCI(Y, X: Integer): Double;
var
  pX, pY : Double;
begin
  if ImageMode = imNormal then
    Result := FGraph[Y, X]
  else
  begin
    if X = 0 then
      pX := (FGraph[Y, 1] - FGraph[Y, 0]) * aHX(X)
    else if X = NumberXm1 then
      pX := (FGraph[Y, NumberXm1] - FGraph[Y, NumberXm1 - 1]) * aHX(X)
    else
      pX := (FGraph[Y, X + 1] - FGraph[Y, X - 1]) * aHX(X);
    if Y = 0 then
      pY := (FGraph[1, X] - FGraph[0, X]) * aHY(Y)
    else if Y = NumberYm1 then
      pY := (FGraph[NumberYm1, X] - FGraph[NumberYm1 - 1, X]) * aHY(Y)
    else
      pY := (FGraph[Y + 1, X] - FGraph[Y - 1, X]) * aHY(Y);
    Result := Hypot(pX, pY);
  end;
end;

function TPcolor.aHX(X: Integer): Double;
begin
  if not IrregularData then Result := FaHX
  else Result := FXGrid2[X];
end;

function TPcolor.aHY(Y: Integer): Double;
begin
  if not IrregularData then Result := FaHY
  else Result := FYGrid2[Y];
end;

procedure TPcolor.SetSizeGraph(AGraphSizeX, AGraphSizeY : Integer;const XMin,XMax,YMin,YMax : Double);
begin
  ChangeGraph := True;
  IrregularData := False;
  if (NumberX <> AGraphSizeX) or (NumberY <> AGraphSizeY) then
  begin
    if AGraphSizeX < 2 then NumberX := 2 else NumberX := AGraphSizeX;
    if AGraphSizeY < 2 then NumberY := 2 else NumberY := AGraphSizeY;
    NumberXm1 := NumberX - 1;
    NumberYm1 := NumberY - 1;
    FGraph := nil;
    SetLength(FGraph,NumberY,NumberX);
    FXGrid := nil;
    FYGrid := nil;
    FXGrid2 := nil;
    FYGrid2 := nil;
    FMaxValGraph := 0;
    FMinValGraph := 0;
    FMaxValGraphX := 0;
    FMaxValGraphY := 0;
    FMinValGraphX := 0;
    FMinValGraphY := 0;
    FGMaxValGraph := 0;
    FGMinValGraph := 0;
    FGMaxValGraphX := 0;
    FGMaxValGraphY := 0;
    FGMinValGraphX := 0;
    FGMinValGraphY := 0;
    MaxMinValGraphValid := True;
    GMaxMinValGraphValid := True;
  end
  else
  begin
    MaxMinValGraphValid := False;
    GMaxMinValGraphValid := False;
  end;

  if (XMin<>FAxes.XMin) or (XMax<>FAxes.XMax)
    or (YMin<>FAxes.YMin) or (YMax<>FAxes.YMax) then
  begin
    if XMin < XMax then
      begin  FAxes.FXMin := XMin;   FAxes.FXMax := XMax;  end
    else
      begin  FAxes.FXMin := XMax;   FAxes.FXMax := XMin;  end;
    if XMin = XMax then
      if XMin > 0 then FAxes.FXMax := XMin * 1.01
      else if XMin < 0 then FAxes.FXMax := XMin * 0.99
      else FAxes.FXMax := 0.01;
    if YMin < YMax then
      begin  FAxes.FYMin := YMin;   FAxes.FYMax := YMax;  end
    else
      begin  FAxes.FYMin := YMax;   FAxes.FYMax := YMin;  end;
    if YMin = YMax then
      if YMin > 0 then FAxes.FYMax := YMin * 1.01
      else if YMin < 0 then FAxes.FYMax := YMin * 0.99
      else FAxes.FYMax := 0.01;
    FaHX := NumberXm1 / (2 * (FAxes.FXMax - FAxes.FXMin));
    FaHY := NumberYm1 / (2 * (FAxes.FYMax - FAxes.FYMin));
    ZoomOut;
  end
  else
  begin
    FaHX := NumberXm1 / (2 * (FAxes.FXMax - FAxes.FXMin));
    FaHY := NumberYm1 / (2 * (FAxes.FYMax - FAxes.FYMin));
    RePaintImage;
  end;
end;

procedure TPcolor.CreateImage;
begin
  ChangeGraph := False;
  case FTypeImage of
    dtFlat          : CreateFlat;
    dtInterp        : CreateInterp;
    dtContour       : CreateContour;
    dtInterpContour : CreateContour;
  end;
end;

procedure TPcolor.CreateFlat;
var
  MaxValGraph2, MinValGraph2 : Double;
  i, j : Integer;
  XStep, YStep, XL, YB, TempVal, mm1, NCm1divMm1 : Double;
  OdivXStep, OdivYStep, LogPrx, LogPry, Logx, Logy, L11, L12, L21, L22 : Double;
  t1,X1,ImageL,ImageR,Y1,ImageT,ImageB,Col,IRmIL : Integer;
begin
  if (NumberX <> 0) and (NumberY <> 0)
   and (Coordinates.ImageWidth <> 0) and (Coordinates.ImageHeight <> 0) then
  begin
    if ImageMode = imNormal then
    begin
       MaxValGraph2 := MaxValGraph;
       MinValGraph2 := MinValGraph;
    end
    else
    begin
       MaxValGraph2 := GMaxValGraph;
       MinValGraph2 := GMinValGraph;
    end;
    mm1 := MaxValGraph2 - MinValGraph2;
    if mm1 <> 0 then NCm1divMm1 := (NumbCol - 1) / mm1
    else NCm1divMm1 := 0;

    if not IrregularData then
    begin
      XStep := HorZoom * FCoordinates.ImageWidth / NumberXm1;
      YStep := VerZoom * FCoordinates.ImageHeight / NumberYm1;

      if (XStep > 1) and (YStep > 1) then
      begin
        XL := LogXToArr(LogLeft) + 0.5;
        YB := LogYToArr(LogBottom) + 0.5;

        X1 := Trunc(XL);
        ImageL := 0;
        ImageR := 0;
        while ImageR < FCoordinates.ImageWidth do
        begin
          X1 := X1 + 1;
          ImageR := Round((X1 - XL) * XStep);
          if ImageR > FCoordinates.ImageWidth then
            ImageR := FCoordinates.ImageWidth;
          Y1 := Trunc(YB);
          ImageT := 0;
          ImageB := 0;
          IRmIL := ImageR - ImageL;
          t1 := Integer(FDIB.Bits) + ImageL;
          while ImageT < FCoordinates.ImageHeight do
          begin
            if not((ImageMode = imGradient) and ((X1 = 1) or (Y1 = 0) or
             (X1 = NumberX) or (Y1 = NumberYm1))) then
            begin
              TempVal := GraphCI(Y1, X1 - 1);
              if (TempVal >= MinValGraph2) then
                if (TempVal <= MaxValGraph2) then
                  Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                else Col := 254
              else Col := 253
            end
            else Col := 0; //fon color
            Y1 := Y1 + 1;
            ImageT := Round((Y1 - YB) * YStep);
            if ImageT > FCoordinates.ImageHeight then
              ImageT := FCoordinates.ImageHeight;
            for i := ImageB to ImageT - 1 do
            begin
              FillMemory(pByte(t1), IRmIL, Col);
              t1 := t1 + FDIB.WordWidth;
            end;
            ImageB := ImageT;
          end;
          ImageL := ImageR;
        end;
      end
      else
      begin
        if Coordinates.ImageWidth < 3 then
          XL := LogXToArr(LogLeft)+0.5
        else
          XL := LogXToArr(LogLeft+(LogRight-LogLeft)/(2*FCoordinates.ImageWidth))+0.5;
        if Coordinates.ImageHeight < 3 then
          YB := LogYToArr(LogBottom)+0.5
        else
          YB := LogYToArr(LogBottom+(LogTop-LogBottom)/(2*FCoordinates.ImageHeight))+0.5;
        OdivXStep := 1 / XStep;
        OdivYStep := 1 / YStep;
        for i := 0 to FCoordinates.ImageHeight - 1 do
        begin
          Y1 := Trunc(YB + i * OdivYStep);
          t1 := Integer(FDIB.Bits) + i * FDIB.WordWidth;
          for j := 0 to FCoordinates.ImageWidth - 1 do
          begin
            X1 := Trunc(XL + j * OdivXStep);
            if not((ImageMode = imGradient) and ((X1 = 0) or (Y1 = 0) or
             (X1 = NumberXm1) or (Y1 = NumberYm1))) then
            begin
              TempVal := GraphCI(Y1, X1);
              if (TempVal >= MinValGraph2) then
                if (TempVal <= MaxValGraph2) then
                  Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                else Col := 254
              else Col := 253
            end
            else Col := 0;
            pByte(t1)^ := Col;
            t1 := t1 + 1;
          end;
        end;
      end;
    end
    else
    begin
      L11 := (LogTop - LogBottom) / FCoordinates.ImageHeight;
      L12 := LogBottom + L11 / 2;
      L21 := (LogRight - LogLeft) / FCoordinates.ImageWidth;
      L22 := LogLeft + L21 / 2;
      Y1 := LogYToYGridFloor(LogBottom);
      LogPry := FYGrid[Y1 + 1] + (FYGrid[Y1 + 2] - FYGrid[Y1 + 1]) / 2;
      for i := 0 to FCoordinates.ImageHeight - 1 do
      begin
        X1 := LogXToXGridFloor(LogLeft);
        LogPrx := FXGrid[X1 + 1] + (FXGrid[X1 + 2] - FXGrid[X1 + 1]) / 2;
        Logy := L12 + i * L11;
        while Logy > LogPry do
        begin
          Y1 := Y1 + 1;
          LogPry := FYGrid[Y1 + 1] + (FYGrid[Y1 + 2] - FYGrid[Y1 + 1]) / 2;
        end;

        TempVal := GraphCI(Y1, X1);
        if not((ImageMode = imGradient) and ((X1 = 0) or (Y1 = 0) or
         (X1 = NumberXm1) or (Y1 = NumberYm1))) then
        begin
          if (TempVal >= MinValGraph2) then
            if (TempVal <= MaxValGraph2) then
              Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
            else Col := 254
          else Col := 253
        end
        else Col := 0;
        t1 := Integer(FDIB.Bits) + i * FDIB.WordWidth;
        for j := 0 to FCoordinates.ImageWidth - 1 do
        begin
          Logx := L22 + j * L21;
          if Logx > LogPrx then
          begin
            repeat
              X1 := X1 + 1;
              LogPrx := FXGrid[X1 + 1] + (FXGrid[X1 + 2] - FXGrid[X1 + 1]) / 2;
            until Logx <= LogPrx;
            TempVal := GraphCI(Y1, X1);
            if not((ImageMode = imGradient) and ((X1 = 0) or (Y1 = 0) or
             (X1 = NumberXm1) or (Y1 = NumberYm1))) then
            begin
              if (TempVal >= MinValGraph2) then
                if (TempVal <= MaxValGraph2) then
                  Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                else Col := 254
              else Col := 253
            end
            else Col := 0;
          end;
          pByte(t1)^ := Col;
          t1 := t1 + 1;
        end;
      end;
    end;
  end;
end;

procedure TPcolor.CreateInterp;
var
  MaxValGraph2, MinValGraph2 : Double;
  XStep, YStep, YAver1, YAver2, XL, YB, TempVal, mm1, NCm1divMm1 : Double;
  OdivXStep, OdivYStep : Double;
  i,j,y1,y2,x1,x2,y2pics,x2pics,Col : Integer;
  t1 : Integer;
  t2,t3, GraphCI11, GraphCI12, LogPrx : Double;
  xcur, ycur, L11, L12, L21, L22 : Double;
begin
  if (NumberX <> 0) and (NumberY <> 0)
   and (Coordinates.ImageWidth <> 0) and (Coordinates.ImageHeight <> 0) then
  begin
    if ImageMode = imNormal then
    begin
       MaxValGraph2 := MaxValGraph;
       MinValGraph2 := MinValGraph;
    end
    else
    begin
       MaxValGraph2 := GMaxValGraph;
       MinValGraph2 := GMinValGraph;
    end;
    mm1 := MaxValGraph2 - MinValGraph2;
    if mm1 <> 0 then NCm1divMm1 := (NumbCol - 1) / mm1
    else NCm1divMm1 := 0;

    if not IrregularData then
    begin
      if Coordinates.ImageWidth < 3 then
        XL := LogXToArr(LogLeft)
      else
        XL := LogXToArr(LogLeft+(LogRight-LogLeft)/(2*FCoordinates.ImageWidth));
      if Coordinates.ImageHeight < 3 then
        YB := LogYToArr(LogBottom)
      else
        YB := LogYToArr(LogBottom+(LogTop-LogBottom)/(2*FCoordinates.ImageHeight));
      XStep := HorZoom * FCoordinates.ImageWidth / NumberXm1;
      YStep := VerZoom * FCoordinates.ImageHeight / NumberYm1;
      OdivXStep := 1 / XStep;
      OdivYStep := 1 / YStep;

      if (XStep > 1) and (YStep > 1) then
      begin
        y1 := Trunc(YB);
        y2 := y1 + 1;
        y2pics := Trunc((y2 - YB) * YStep);
        for i := 0 to FCoordinates.ImageHeight - 1 do
        begin
          x1 := Trunc(XL);
          x2 := x1 + 1;
          x2pics := Trunc((x2 - XL) * XStep);
          if i > y2pics then
          begin
            y1 := y2;
            y2 := y1 + 1;
            y2pics := Trunc((y2 - YB) * YStep);
          end;
          ycur := YB + i * OdivYStep;
          t2 := ycur - y1;
          GraphCI11 := GraphCI(y1,x1);
          GraphCI12 := GraphCI(y1,x2);
          YAver1 := t2 * (GraphCI(y2,x1) - GraphCI11) + GraphCI11;
          YAver2 := t2 * (GraphCI(y2,x2) - GraphCI12) + GraphCI12;
          t1 := Integer(FDIB.Bits) + i * FDIB.WordWidth;
          for j := 0 to FCoordinates.ImageWidth - 1 do
          begin
            if j > x2pics then
            begin
              x1 := x2;
              x2 := x1 + 1;
              x2pics := Trunc((x2 - XL) * XStep);
              YAver1 := YAver2;
              GraphCI12 := GraphCI(y1,x2);
              YAver2 := t2 * (GraphCI(y2,x2) - GraphCI12) + GraphCI12;
            end;
            if not((ImageMode = imGradient) and ((x1 = 0) or (y1 = 0) or
             (x2 = NumberXm1) or (y2 = NumberYm1))) then
            begin
              xcur := XL + j * OdivXStep;
              t3 := xcur - x1;
              TempVal := t3 * (YAver2 - YAver1) + YAver1;
              if (TempVal >= MinValGraph2) then
                if (TempVal <= MaxValGraph2) then
                  Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                else Col := 254
              else Col := 253
            end
            else Col := 0;
            pByte(t1)^ := Col;
            t1 := t1 + 1;
          end;
        end;
      end
      else
      for i := 0 to FCoordinates.ImageHeight - 1 do
      begin
        ycur := YB + i * OdivYStep;
        y1 := Trunc(ycur);
        y2 := y1 + 1;
        t1 := Integer(FDIB.Bits) + i * FDIB.WordWidth;
        t2 := ycur - y1;
        for j := 0 to FCoordinates.ImageWidth - 1 do
        begin
          xcur := XL + j * OdivXStep;
          x1 := Trunc(xcur);
          x2 := x1 + 1;
          if not((ImageMode = imGradient) and ((x1 = 0) or (y1 = 0) or
           (x2 = NumberXm1) or (y2 = NumberYm1))) then
          begin
            GraphCI11 := GraphCI(y1,x1);
            GraphCI12 := GraphCI(y1,x2);
            YAver1 := t2 * (GraphCI(y2,x1) - GraphCI11) + GraphCI11;
            t3 := xcur - x1;
            TempVal := t3*(t2*(GraphCI(y2,x2)-GraphCI12)+GraphCI12-YAver1)+YAver1;
            if (TempVal >= MinValGraph2) then
              if (TempVal <= MaxValGraph2) then
                Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
              else Col := 254
            else Col := 253
          end
          else Col := 0;
          pByte(t1)^ := Col;
          t1 := t1 + 1;
        end;
      end;
    end
    else
    begin
      L11 := (LogTop - LogBottom) / FCoordinates.ImageHeight;
      L12 := LogBottom + L11 / 2;
      L21 := (LogRight - LogLeft) / FCoordinates.ImageWidth;
      L22 := LogLeft + L21 / 2;
      y2 := 0;
      for i := 0 to FCoordinates.ImageHeight - 1 do
      begin
        ycur := L12 + i * L11;
        while (FYGrid[y2 + 1] < ycur) and (y2 < NumberYm1) do y2 := y2 + 1;
        y1 := y2 - 1;
        t1 := Integer(FDIB.Bits) + i * FDIB.WordWidth;
        t2 := (ycur - FYGrid[y1 + 1]) / (FYGrid[y2 + 1] - FYGrid[y1 + 1]);
        x1 := LogXToXGridFloor(LogLeft);
        x2 := x1 + 1;
        LogPrx := FXGrid[x2 + 1];
        GraphCI11 := GraphCI(y1,x1);
        GraphCI12 := GraphCI(y1,x2);
        YAver1 := t2 * (GraphCI(y2,x1) - GraphCI11) + GraphCI11;
        YAver2 := t2 * (GraphCI(y2,x2) - GraphCI12) + GraphCI12;
        for j := 0 to FCoordinates.ImageWidth - 1 do
        begin
          xcur := L22 + j * L21;
          if xcur > LogPrx then
          begin
            while (LogPrx < xcur) and (x2 < NumberXm1) do
            begin
              x2 := x2 + 1;  
              LogPrx := FXGrid[x2 + 1];
            end;
            x1 := x2 - 1;
            GraphCI11 := GraphCI(y1,x1);
            GraphCI12 := GraphCI(y1,x2);
            YAver1 := t2 * (GraphCI(y2,x1) - GraphCI11) + GraphCI11;
            YAver2 := t2 * (GraphCI(y2,x2) - GraphCI12) + GraphCI12;
          end;
          if not((ImageMode = imGradient) and ((x1 = 0) or (y1 = 0) or
           (x2 = NumberXm1) or (y2 = NumberYm1))) then
          begin
            t3 := (xcur - FXGrid[x1 + 1]) / (LogPrx - FXGrid[x1 + 1]);
            TempVal := t3 * (YAver2 - YAver1) + YAver1;
            if (TempVal >= MinValGraph2) then
              if (TempVal <= MaxValGraph2) then
                Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
              else Col := 254
            else Col := 253
          end
          else Col := 0;
          pByte(t1)^ := Col;
          t1 := t1 + 1;
        end;
      end;
    end;
  end;
end;

procedure TPcolor.CreateContour;
var
  MaxValGraph2, MinValGraph2 : Double;
  XStep, YStep, YAver11, YAver12, mm1, NCm1divMm1 : Double;
  OdivXStep, OdivYStep : Double;
  a11, a12, a21, a22 : Double;
  hh, XL, YB : Double;
  Colors : array of Byte;
  PixRow : array of Double;
  y11, y12, y12pics, x21, x22, x22pics, temp : Integer;
  i, j, t1, ContLinP1 : Integer;
  ycur, xcur, t2, t3 : Double;
  L1, L2, GraphCI1121, GraphCI1122, LogPrx : Double;
  qn, Col : Integer;
  Max3, Min3, TempVal : Double;

begin
  if (NumberX <> 0) and (NumberY <> 0)
   and (Coordinates.ImageWidth <> 0) and (Coordinates.ImageHeight <> 0) then
  begin
    ContLinP1 := FContourLines + 1;
    SetLength(Colors,FContourLines + 2);
    for i := 0 to ContLinP1 do
      Colors[i] := Round((NumbCol - 1) * i / ContLinP1) + 1;

    if ImageMode = imNormal then
    begin
       MaxValGraph2 := MaxValGraph;
       MinValGraph2 := MinValGraph;
    end
    else
    begin
       MaxValGraph2 := GMaxValGraph;
       MinValGraph2 := GMinValGraph;
    end;
    mm1 := MaxValGraph2 - MinValGraph2;
    if mm1 <> 0 then
    begin
      NCm1divMm1 := (NumbCol - 1) / mm1;
      hh := ContLinP1 / mm1;
    end
    else
    begin
      NCm1divMm1 := 0;
      hh := 0;
    end;

    SetLength(PixRow, FCoordinates.ImageWidth + 1);

    FDIB.Clear;

    if not IrregularData then
    begin
      XL := LogXToArr(LogLeft);
      YB := LogYToArr(LogBottom);
      XStep := HorZoom * FCoordinates.ImageWidth / NumberXm1;
      YStep := VerZoom * FCoordinates.ImageHeight / NumberYm1;
      OdivXStep := 1 / XStep;
      OdivYStep := 1 / YStep;

      if (XStep > 1) and (YStep > 1) then
      begin
        y11 := Trunc(YB);
        y12 := y11 + 1;
        y12pics := Trunc((y12 - YB) * YStep);
        x21 := Trunc(XL);
        x22 := x21 + 1;
        x22pics := Trunc((x22 - XL) * XStep);
        t2 := YB - y11;
        GraphCI1121 := GraphCI(y11,x21);
        GraphCI1122 := GraphCI(y11,x22);
        YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
        YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1121;
        for j := 0 to FCoordinates.ImageWidth do
        begin
          if (j > x22pics) and (x22 < NumberXm1) then
          begin
            x21 := x22;
            x22 := x22 + 1;
            x22pics := Trunc((x22 - XL) * XStep);
            YAver11 := YAver12;
            GraphCI1122 := GraphCI(y11,x22);
            YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
          end;
          t3 := j * OdivXStep + XL - x21;
          PixRow[j] := t3 * (YAver12 - YAver11) + YAver11;
        end;
        for i := 1 to FCoordinates.ImageHeight do
        begin
          x21 := Trunc(XL);
          x22 := x21 + 1;
          x22pics := Trunc((x22 - XL) * XStep);
          if i > y12pics then
            if y12 < NumberYm1 then
            begin
              y11 := y12;
              y12 := y12 + 1;
              y12pics := Trunc((y12 - YB) * YStep);
            end;
          ycur := YB + i * OdivYStep;
          t2 := ycur - y11;
          GraphCI1121 := GraphCI(y11,x21);
          GraphCI1122 := GraphCI(y11,x22);
          YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
          YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
          t3 := XL - x21;
          a21 := t3 * (YAver12 - YAver11) + YAver11;
          a11 := PixRow[0];
          PixRow[0] := a21;
          t1 := Integer(FDIB.Bits) + (i - 1) * FDIB.WordWidth - 1;
          for j := 1 to FCoordinates.ImageWidth do
          begin
            if (j > x22pics) and (x22 < NumberXm1) then
            begin
              x21 := x22;
              x22 := x22 + 1;
              x22pics := Trunc((x22 - XL) * XStep);
              YAver11 := YAver12;
              GraphCI1122 := GraphCI(y11,x22);
              YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
            end;
            xcur := XL + j * OdivXStep;
            t3 := xcur - x21;
            a22 := t3 * (YAver12 - YAver11) + YAver11;
            a12 := PixRow[j];
            PixRow[j] := a22;

            if not((ImageMode = imGradient) and ((x21 = 0) or (y11 = 0) or
             (x22 = NumberXm1) or (y12 = NumberYm1))) then
            begin
              Max3 := a11;
              Min3 := a11;
              if a12 > Max3 then Max3 := a12;
              if a21 > Max3 then Max3 := a21;
              if a22 > Max3 then Max3 := a22;
              if a12 < Min3 then Min3 := a12;
              if a21 < Min3 then Min3 := a21;
              if a22 < Min3 then Min3 := a22;

              if not (FTypeImage = dtInterpContour) then
              begin
                if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
                begin
                  qn := Trunc((Max3 - MinValGraph2) * hh);
                  if qn > ContLinP1 then qn := ContLinP1;
                  if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)){ceil} then
                    pByte(t1 + j)^ := Colors[qn];
                end
              end
              else
              begin
                if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
                begin
                  qn := Trunc((Max3 - MinValGraph2) * hh);
                  if qn > ContLinP1 then qn := ContLinP1;
                  if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)) then
                    Col := 255
                  else
                  begin
                    TempVal := (a11 + a12 + a21 + a22) / 4;
                    if (TempVal >= MinValGraph2) then
                      if (TempVal <= MaxValGraph2) then
                        Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                      else Col := 254
                    else Col := 253;
                  end;
                end
                else
                begin
                  TempVal := (a11 + a12 + a21 + a22) / 4;
                  if (TempVal >= MinValGraph2) then
                    if (TempVal <= MaxValGraph2) then
                      Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                    else Col := 254
                  else Col := 253;
                end;
                pByte(t1 + j)^ := Col;
              end;
            end;
            a11 := a12;
            a21 := a22;
          end;
        end;
      end
      else
      begin
        y11 := Trunc(YB);
        y12 := y11 + 1;
        t2 := YB - y11;
        for j := 0 to FCoordinates.ImageWidth - 1 do
        begin
          x21 := Trunc(XL + j * OdivXStep);
          x22 := x21 + 1;
          GraphCI1121 := GraphCI(y11,x21);
          GraphCI1122 := GraphCI(y11,x22);
          YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
          t3 := j * OdivXStep - x21 + XL;
          PixRow[j] := t3*(t2*(GraphCI(y12,x22)-GraphCI1122)+GraphCI1122-YAver11)+YAver11;
        end;
        temp := Trunc(XL + FCoordinates.ImageWidth * OdivXStep);
        if temp < NumberXm1 then x21 := temp
        else x21 := NumberXm1 - 1;
        x22 := x21 + 1;
        YAver11 := t2*(GraphCI(y12,x21)-GraphCI(y11,x21))+GraphCI(y11,x21);
        t3 := FCoordinates.ImageWidth * OdivXStep - x21 + XL;
        PixRow[FCoordinates.ImageWidth] := t3*(t2*(GraphCI(y12,x22)-GraphCI(y11,x22))
         + GraphCI(y11,x22) - YAver11) + YAver11;
        for i := 1 to FCoordinates.ImageHeight do
        begin
          ycur := YB + i * OdivYStep;
          temp := Trunc(ycur);
          if temp < NumberYm1 then y11 := temp
          else y11 := NumberYm1 - 1;
          y12 := y11 + 1;
          x21 := Trunc(XL);
          x22 := x21 + 1;
          t2 := ycur - y11;
          GraphCI1121 := GraphCI(y11,x21);
          GraphCI1122 := GraphCI(y11,x22);
          YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
          t3 := XL + x21;
          a21 := t3*(t2*(GraphCI(y12,x22)-GraphCI1122)+GraphCI1122-YAver11)+YAver11;
          a11 := PixRow[0];
          PixRow[0] := a21;
          t1 := Integer(FDIB.Bits) + (i - 1) * FDIB.WordWidth - 1;
          for j := 1 to FCoordinates.ImageWidth do
          begin
            xcur := XL + j * OdivXStep;
            temp := Trunc(xcur);
            if temp < NumberXm1 then x21 := temp
            else x21 := NumberXm1 - 1;
            x22 := x21 + 1;
            GraphCI1121 := GraphCI(y11,x21);
            GraphCI1122 := GraphCI(y11,x22);
            YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
            t3 := xcur - x21;
            a22 := t3 * (t2 * (GraphCI(y12,x22) - GraphCI1122)
             + GraphCI1122 - YAver11) + YAver11;
            a12 := PixRow[j];
            PixRow[j] := a22;

            if not((ImageMode = imGradient) and ((x21 = 0) or (y11 = 0) or
             (x22 = NumberXm1) or (y12 = NumberYm1))) then
            begin
              Max3 := a11;
              Min3 := a11;
              if a12 > Max3 then Max3 := a12;
              if a21 > Max3 then Max3 := a21;
              if a22 > Max3 then Max3 := a22;
              if a12 < Min3 then Min3 := a12;
              if a21 < Min3 then Min3 := a21;
              if a22 < Min3 then Min3 := a22;

              if not (FTypeImage = dtInterpContour) then
              begin
                if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
                begin
                  qn := Trunc((Max3 - MinValGraph2) * hh);
                  if qn > ContLinP1 then qn := ContLinP1;
                  if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)){ceil} then
                    pByte(t1 + j)^ := Colors[qn];
                end
              end
              else
              begin
                if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
                begin
                  qn := Trunc((Max3 - MinValGraph2) * hh);
                  if qn > ContLinP1 then qn := ContLinP1;
                  if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)) then
                    Col := 255
                  else
                  begin
                    TempVal := (a11 + a12 + a21 + a22) / 4;
                    if (TempVal >= MinValGraph2) then
                      if (TempVal <= MaxValGraph2) then
                        Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                      else Col := 254
                    else Col := 253;
                  end;
                end
                else
                begin
                  TempVal := (a11 + a12 + a21 + a22) / 4;
                  if (TempVal >= MinValGraph2) then
                    if (TempVal <= MaxValGraph2) then
                      Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                    else Col := 254
                  else Col := 253;
                end;
                pByte(t1 + j)^ := Col;
              end;
            end;
            a11 := a12;
            a21 := a22;
          end;
        end;
      end;
    end
    else
    begin
      L1 := (LogTop - LogBottom) / FCoordinates.ImageHeight;
      L2 := (LogRight - LogLeft) / FCoordinates.ImageWidth;

      y11 := LogYToYGridFloor(LogBottom);
      y12 := y11 + 1;
      t2 := (LogBottom - FYGrid[y11 + 1]) / (FYGrid[y12 + 1] - FYGrid[y11 + 1]);
      x21 := LogXToXGridFloor(LogLeft);
      x22 := x21 + 1;
      LogPrx := FXGrid[x22 + 1];
      GraphCI1121 := GraphCI(y11,x21);
      GraphCI1122 := GraphCI(y11,x22);
      YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
      YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
      for j := 0 to FCoordinates.ImageWidth do
      begin
        xcur := LogLeft + j * L2;
        if xcur > LogPrx then
        begin
          while (LogPrx < xcur) and (x22 < NumberXm1) do
          begin
            x22 := x22 + 1;
            LogPrx := FXGrid[x22 + 1];
          end;
          x21 := x22 - 1;
          GraphCI1121 := GraphCI(y11,x21);
          GraphCI1122 := GraphCI(y11,x22);
          YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
          YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
        end;
        t3 := (xcur - FXGrid[x21 + 1]) / (LogPrx - FXGrid[x21 + 1]);
        PixRow[j] := t3 * (YAver12 - YAver11) + YAver11;
      end;
      for i := 1 to FCoordinates.ImageHeight do
      begin
        ycur := LogBottom + i * L1;
        while (FYGrid[y12 + 1] < ycur) and (y12 < NumberYm1) do y12 := y12 + 1;
        y11 := y12 - 1;
        t2 := (ycur - FYGrid[y11 + 1]) / (FYGrid[y12 + 1] - FYGrid[y11 + 1]);
        x21 := LogXToXGridFloor(LogLeft);
        x22 := X21 + 1;
        LogPrx := FXGrid[x22 + 1];
        GraphCI1121 := GraphCI(y11,x21);
        GraphCI1122 := GraphCI(y11,x22);
        YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
        YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
        t3 := (LogLeft - FXGrid[x21 + 1]) / (LogPrx - FXGrid[x21 + 1]);
        a21 := t3 * (YAver12 - YAver11) + YAver11;
        a11 := PixRow[0];
        PixRow[0] := a21;
        t1 := Integer(FDIB.Bits) + (i - 1) * FDIB.WordWidth - 1;
        for j := 1 to FCoordinates.ImageWidth do
        begin
          xcur := LogLeft + j * L2;
          if xcur > LogPrx then
          begin
            while (LogPrx < xcur) and (x22 < NumberXm1) do
            begin
              x22 := x22 + 1;
              LogPrx := FXGrid[x22 + 1];
            end;
            x21 := x22 - 1;
            GraphCI1121 := GraphCI(y11,x21);
            GraphCI1122 := GraphCI(y11,x22);
            YAver11 := t2 * (GraphCI(y12,x21) - GraphCI1121) + GraphCI1121;
            YAver12 := t2 * (GraphCI(y12,x22) - GraphCI1122) + GraphCI1122;
          end;
          t3 := (xcur - FXGrid[x21 + 1]) / (LogPrx - FXGrid[x21 + 1]);
          a22 := t3 * (YAver12 - YAver11) + YAver11;
          a12 := PixRow[j];
          PixRow[j] := a22;

          if not((ImageMode = imGradient) and ((x21 = 0) or (y11 = 0) or
           (x22 = NumberXm1) or (y12 = NumberYm1))) then
          begin
            Max3 := a11;
            Min3 := a11;
            if a12 > Max3 then Max3 := a12;
            if a21 > Max3 then Max3 := a21;
            if a22 > Max3 then Max3 := a22;
            if a12 < Min3 then Min3 := a12;
            if a21 < Min3 then Min3 := a21;
            if a22 < Min3 then Min3 := a22;
            if not (FTypeImage = dtInterpContour) then
            begin
              if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
              begin
                qn := Trunc((Max3 - MinValGraph2) * hh);
                if qn > ContLinP1 then qn := ContLinP1;
                if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)){ceil} then
                  pByte(t1 + j)^ := Colors[qn];
              end
            end
            else
            begin
              if (Max3 >= MinValGraph2) and (Min3 <= MaxValGraph2) then
              begin
                qn := Trunc((Max3 - MinValGraph2) * hh);
                if qn > ContLinP1 then qn := ContLinP1;
                if qn >= (MaxInt-Trunc(MaxInt-(Min3-MinValGraph2)*hh)) then
                  Col := 255
                else
                begin
                  TempVal := (a11 + a12 + a21 + a22) / 4;
                  if (TempVal >= MinValGraph2) then
                    if (TempVal <= MaxValGraph2) then
                      Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                    else Col := 254
                  else Col := 253;
                end;
              end
              else
              begin
                TempVal := (a11 + a12 + a21 + a22) / 4;
                if (TempVal >= MinValGraph2) then
                  if (TempVal <= MaxValGraph2) then
                    Col := Round((TempVal - MinValGraph2) * NCm1divMm1) + 1
                  else Col := 254
                else Col := 253;
              end;
              pByte(t1 + j)^ := Col;
            end;
          end;
          a11 := a12;
          a21 := a22;
        end;
      end;
    end;
  end;
end;

function TPcolor.ValueInPosLog(APosLogX, APosLogY : Double): Double;
var
  YAver1, XX, YY : Double;
  y1,y2,x1,x2 : Integer;
  t2,t3 : Double;
begin
  if (NumberX <> 0) and (NumberY <> 0) then
  begin
    PosLogToPosArr(APosLogX, APosLogY, XX, YY);
    if FTypeImage=dtFlat then
    begin
      XX := Trunc(XX + 0.5);
      YY := Trunc(YY + 0.5);
    end;
    y1 := Trunc(YY);
    if y1 >= NumberYm1 then y1 := NumberYm1 - 1;
    y2 := y1 + 1;
    t2 := YY - y1;
    x1 := Trunc(XX);
    if x1 >= NumberXm1 then x1 := NumberXm1 - 1;
    x2 := x1 + 1;
    YAver1 := t2 * (GraphCI(y2,x1) - GraphCI(y1,x1)) + GraphCI(y1,x1);
    t3 := XX - x1;
    Result := t3 * (t2 * (GraphCI(y2,x2) - GraphCI(y1,x2))
      + GraphCI(y1,x2) - YAver1) + YAver1
  end
  else
    Result := 0
end;

procedure TPcolor.SetAbout;
begin
end;

procedure TPcolor.SetDefaultData;
begin
  NumberX := 0;
  NumberY := 0;
  FGraph := nil;
  FXGrid := nil;
  FYGrid := nil;
  FXGrid2 := nil;
  FYGrid2 := nil;
  FDIB.Clear;
  FMaxValGraph := 0;
  FMinValGraph := 0;
  FMaxValGraphX := 0;
  FMaxValGraphY := 0;
  FMinValGraphX := 0;
  FMinValGraphY := 0;
  FGMaxValGraph := 0;
  FGMinValGraph := 0;
  FGMaxValGraphX := 0;
  FGMaxValGraphY := 0;
  FGMinValGraphX := 0;
  FGMinValGraphY := 0;
  MaxMinValGraphValid := True;
  GMaxMinValGraphValid := True;
  ChangeGraph := True;
  IrregularData := False;
  ZoomOut;
  RePaintImage;
end;

procedure TPcolor.SetDrawFacet(ADrawFacet : boolean);
begin
  FDrawFacet := ADrawFacet;
  RePaintImage;
end;

procedure TPcolor.SetColorFacet(AColor : TColor);
begin
  FColorFacet := AColor;
  if FDrawFacet then RePaintImage;
end;

procedure TPcolor.SetColorFon(AColor : TColor);
begin
  FColorFon := AColor;
  Paint;
end;

procedure TPcolor.SetColorImageFon(AColor : TColor);
begin
  FColorImageFon := AColor;
  ChangePalette := True;
  RePaintImage;
end;

procedure TPcolor.SetColorContour(AColor : TColor);
begin
  if FColorContour <> AColor then
  begin
    FColorContour := AColor;
    if TypeImage = dtInterpContour then
    begin
      ChangePalette := True;
      RePaintImage;
    end;
  end;
end;

procedure TPcolor.SetColorHigher(AColor : TColor);
begin
  if FColorHigher <> AColor then
  begin
    FColorHigher := AColor;
    if TypeImage <> dtContour then
    begin
      ChangePalette := True;
      Paint;
    end;
  end;
end;

procedure TPcolor.SetColorLower(AColor : TColor);
begin
  if FColorLower <> AColor then
  begin
    FColorLower := AColor;
    if TypeImage <> dtContour then
    begin
      ChangePalette := True;
      Paint;
    end;
  end;
end;

procedure TPcolor.SetColorMap(AColorMap : TColorMaps);
begin
  FColorMap := AColorMap;
  ChangePalette := True;
  Paint;
end;

procedure TPcolor.SetColorMaxValue(AColor : TColor);
begin
  FColorMaxValue := AColor;
  if FColorMap = cmLine then
  begin
    ChangePalette := True;
    Paint;
  end;
end;

procedure TPcolor.SetColorMinValue(AColor : TColor);
begin
  FColorMinValue := AColor;
  if FColorMap=cmLine then
  begin
    ChangePalette := True;
    Paint;
  end;  
end;

procedure TPcolor.SetContourLines(AContourLines : Integer);
begin
  if AContourLines<2 then FContourLines:=1
  else FContourLines := AContourLines;
  if FTypeImage = dtContour then
  begin
    CreateContour;
    RePaintImage;
  end;  
end;

procedure TPcolor.SetTypeImage(ATypeImage : TDrawTypes);
begin
  if (FTypeImage <> ATypeImage) then
  begin
    FTypeImage := ATypeImage;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

procedure TPcolor.SetImageMode(AImageMode : TImageModes);
begin
  if (FImageMode <> AImageMode) then
  begin
    FImageMode := AImageMode;
    ChangeGraph := True;
    RePaintImage;
  end;
end;

procedure TPcolor.CreatePaletta;
var
   CMinRed,CMinGreen,CMinBlue : byte;
   CMaxRed,CMaxGreen,CMaxBlue : byte;
   i : integer;
begin
  ChangePalette := False;
  case FColorMap of
   cmLine :
    begin
      CMinRed := ColorToRGB(FColorMinValue) and $000000FF;
      CMinGreen := (ColorToRGB(FColorMinValue) and $0000FF00) shr 8;
      CMinBlue := (ColorToRGB(FColorMinValue) and $00FF0000) shr 16;
      CMaxRed := ColorToRGB(FColorMaxValue) and $000000FF;
      CMaxGreen := (ColorToRGB(FColorMaxValue) and $0000FF00) shr 8;
      CMaxBlue := (ColorToRGB(FColorMaxValue) and $00FF0000) shr 16;
      for i:=1 to NumbCol do
        begin
          FLogPalette.palEntry[i].peRed := Round((i-1)*(CMaxRed-CMinRed)/(NumbCol-1))+CMinRed;
          FLogPalette.palEntry[i].peGreen := Round((i-1)*(CMaxGreen-CMinGreen)/(NumbCol-1))+CMinGreen;
          FLogPalette.palEntry[i].peBlue := Round((i-1)*(CMaxBlue-CMinBlue)/(NumbCol-1))+CMinBlue;
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
    end;
   cmSpectr :
    begin                   
      for i:=    1 to Trunc(NumbCol/8*1) do
        begin
          FLogPalette.palEntry[i].peRed   := 0;
          FLogPalette.palEntry[i].peGreen := 0;
          FLogPalette.palEntry[i].peBlue  := 127+Round(255*i*4/NumbCol);
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
      for i:=Trunc(NumbCol/8*1)+1 to Trunc(NumbCol/8*3) do
        begin
          FLogPalette.palEntry[i].peRed   := 0;
          FLogPalette.palEntry[i].peGreen := Round(255*(i-NumbCol/8)*4/NumbCol);
          FLogPalette.palEntry[i].peBlue  := 255;
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
      for i:=Trunc(NumbCol/8*3)+1 to Trunc(NumbCol/8*5) do
        begin
          FLogPalette.palEntry[i].peRed   := Round(255*(i-3*NumbCol/8)*4/NumbCol);
          FLogPalette.palEntry[i].peGreen := 255;
          FLogPalette.palEntry[i].peBlue  := Round(255*(5*NumbCol/8-i)*4/NumbCol);
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
      for i:=Trunc(NumbCol/8*5)+1 to Trunc(NumbCol/8*7) do
        begin
          FLogPalette.palEntry[i].peRed   := 255;
          FLogPalette.palEntry[i].peGreen := round(255*(7*NumbCol/8-i)*4/NumbCol);
          FLogPalette.palEntry[i].peBlue  := 0;
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
      for i:=trunc(NumbCol/8*7)+1 to NumbCol do
        begin
          FLogPalette.palEntry[i].peRed   := round(255*(9*NumbCol/8-i)*4/NumbCol);
          FLogPalette.palEntry[i].peGreen := 0;
          FLogPalette.palEntry[i].peBlue  := 0;
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
    end;
   cmRandom :
    begin
      Randomize;
      for i := 1 to NumbCol do
        begin
          FLogPalette.palEntry[i].peRed   := Random(255);
          FLogPalette.palEntry[i].peGreen := Random(255);
          FLogPalette.palEntry[i].peBlue  := Random(255);
          FLogPalette.palEntry[i].peFlags := pc_Reserved;
        end;
    end;
   cmUser :
    begin
      for i := 1 to NumbCol do
        FLogPalette.palEntry[i] := FUserPalette.palEntry[i];
    end;
  end;
  //fon color
  FLogPalette.palEntry[0].peRed := ColorToRGB(FColorImageFon) and $000000FF;
  FLogPalette.palEntry[0].peGreen := (ColorToRGB(FColorImageFon) and $0000FF00) shr 8;
  FLogPalette.palEntry[0].peBlue := (ColorToRGB(FColorImageFon) and $00FF0000) shr 16;
  FLogPalette.palEntry[0].peFlags := pc_Reserved;
  FLogPalette.palEntry[253].peRed := ColorToRGB(FColorLower) and $000000FF;
  FLogPalette.palEntry[253].peGreen := (ColorToRGB(FColorLower) and $0000FF00) shr 8;
  FLogPalette.palEntry[253].peBlue := (ColorToRGB(FColorLower) and $00FF0000) shr 16;
  FLogPalette.palEntry[253].peFlags := pc_Reserved;
  FLogPalette.palEntry[254].peRed := ColorToRGB(FColorHigher) and $000000FF;
  FLogPalette.palEntry[254].peGreen := (ColorToRGB(FColorHigher) and $0000FF00) shr 8;
  FLogPalette.palEntry[254].peBlue := (ColorToRGB(FColorHigher) and $00FF0000) shr 16;
  FLogPalette.palEntry[254].peFlags := pc_Reserved;
  FLogPalette.palEntry[255].peRed := ColorToRGB(FColorContour) and $000000FF;
  FLogPalette.palEntry[255].peGreen := (ColorToRGB(FColorContour) and $0000FF00) shr 8;
  FLogPalette.palEntry[255].peBlue := (ColorToRGB(FColorContour) and $00FF0000) shr 16;
  FLogPalette.palEntry[255].peFlags := pc_Reserved;
  FDIB.SetPalette(FLogPalette);
end;

function TPColor.IsPosInImage(APosImageX, APosImageY : Integer): Boolean;
begin
  if (APosImageX < FCoordinates.ImageLeft)
     or (APosImageY < FCoordinates.ImageTop)
     or (APosImageX > FCoordinates.ImageLeft + FCoordinates.ImageWidth - 1)
     or (APosImageY > FCoordinates.ImageTop + FCoordinates.ImageHeight - 1)
  then Result := False
  else Result := True;
end;

function TPColor.PosImageToPosLog(APosImageX, APosImageY : Integer;
            var XAx, YAx: Double): Boolean;
begin
  Result := IsPosInImage(APosImageX, APosImageY);
  if (APosImageX < FCoordinates.ImageLeft)
    then APosImageX := FCoordinates.ImageLeft;
  if (APosImageY < FCoordinates.ImageTop)
    then APosImageY := FCoordinates.ImageTop;
  if (APosImageX > FCoordinates.ImageLeft + FCoordinates.ImageWidth - 1)
    then APosImageX := FCoordinates.ImageLeft + FCoordinates.ImageWidth - 1;
  if (APosImageY > FCoordinates.ImageTop + FCoordinates.ImageHeight - 1)
    then APosImageY := FCoordinates.ImageTop + FCoordinates.ImageHeight - 1;
  if FCoordinates.ImageWidth = 1 then XAx := Axes.XMin
  else XAx := LogLeft+(LogRight-LogLeft)*(APosImageX-FCoordinates.ImageLeft)/(FCoordinates.ImageWidth-1);
  if FCoordinates.ImageHeight = 1 then YAx := Axes.YMin
  else YAx := LogTop-(LogTop-LogBottom)*(APosImageY-FCoordinates.ImageTop)/(FCoordinates.ImageHeight-1);
end;

function TPColor.LogXToXGridCeil(Value : Double) : Integer;
var
  i : integer;
begin
  i := 1;
  while FXGrid[i] < Value do i := i + 1;
  Result := i - 1;
end;

function TPColor.LogXToXGridFloor(Value : Double) : Integer;
var
  i : integer;
begin
  i := LogXToXGridCeil(Value);
  if FXGrid[i + 1] = Value then Result := i
  else Result := i - 1;
end;

function TPColor.LogYToYGridCeil(Value : Double) : Integer;
var
  i : integer;
begin
  i := 1;
  while FYGrid[i] < Value do i := i + 1;
  Result := i - 1;
end;

function TPColor.LogYToYGridFloor(Value : Double) : Integer;
var
  i : integer;
begin
  i := LogYToYGridCeil(Value);
  if FYGrid[i + 1] = Value then Result := i
  else Result := i - 1;
end;

function TPColor.LogXToArr(Value : Double) : Double;
var
  i : integer;
begin
  if not IrregularData then
    Result := ((Value-FAxes.XMin) / (FAxes.XMax-FAxes.XMin)) * NumberXm1
  else
  begin
    i := LogXToXGridFloor(Value);
    Result := i + (Value - FXGrid[i+1]) / (FXGrid[i+2] - FXGrid[i+1]);
  end;
end;

function TPColor.LogYToArr(Value : Double) : Double;
var
  i : integer;
begin
  if not IrregularData then
    Result := ((Value-FAxes.YMin) / (FAxes.YMax-FAxes.YMin)) * NumberYm1
  else
  begin
    i := LogYToYGridFloor(Value);
    Result := i + (Value - FYGrid[i+1]) / (FYGrid[i+2] - FYGrid[i+1]);
  end;
end;

function TPColor.PosLogToPosArr(ALogX, ALogY : Double;
            var ArrX, ArrY: Double) : Boolean;
begin
  Result := True;
  if (ALogX < Axes.XMin) then
  begin
    ALogX := Axes.XMin;
    Result := False
  end;
  if (ALogX > Axes.XMax) then
  begin
    ALogX := Axes.XMax;
    Result := False
  end;
  if (ALogY < Axes.YMin) then
  begin
    ALogY := Axes.YMin;
    Result := False
  end;
  if (ALogY > Axes.YMax) then
  begin
    ALogY := Axes.YMax;
    Result := False
  end;
  ArrX := LogXToArr(ALogX);
  ArrY := LogYToArr(ALogY);
end;

function TPColor.PosArrToPosLog(AArrX, AArrY : Double;
            var LogX, LogY: Double) : Boolean;
var
  i : integer;
begin
  Result := True;
  if AArrX < 0 then
  begin
    AArrX := 0;
    Result := False
  end;
  if AArrX > NumberXm1 then
  begin
    AArrX := NumberXm1;
    Result := False
  end;
  if AArrY < 0 then
  begin
    AArrY := 0;
    Result := False
  end;
  if AArrY > NumberYm1 then
  begin
    AArrY := NumberYm1;
    Result := False
  end;
  if not IrregularData then
  begin
    LogX := (AArrX / NumberXm1)*(FAxes.XMax-FAxes.XMin)+FAxes.XMin;
    LogY := (AArrY / NumberYm1)*(FAxes.YMax-FAxes.YMin)+FAxes.YMin;
  end
  else
  begin
    i := Trunc(AArrX);
    LogX := FXGrid[i+1] + (AArrX - i) * (FXGrid[i+2] - FXGrid[i+1]);
    i := Trunc(AArrY);
    LogY := FYGrid[i+1] + (AArrY - i) * (FYGrid[i+2] - FYGrid[i+1]);
  end;
end;

procedure TPColor.TitleChanged(Sender: TObject);
begin
  Sizes;
  Paint;
end;

procedure TPcolor.SetTitle(Value: TTitle);
begin
  FTitle.Assign(Value);
end;

procedure TPcolor.DoDrawText(var Rect: TRect; Flags: Longint; Canva: TCanvas);
var
  Text : String;
begin
  Text := FTitle.Text;
  if (Flags and DT_CALCRECT <> 0) and (Text = '') then Text := ' ';
  Flags := Flags or DT_NOPREFIX;
  Canva.Font := FTitle.Font;
  DrawText(Canva.Handle, PChar(Text), Length(Text), Rect, Flags);
end;

procedure TPcolor.SetAxes(Value: TAxes);
begin
  FAxes.Assign(Value);
end;

procedure TPcolor.Zoom;
begin
  if Assigned(FOnZoom) then FOnZoom(Self);
end;

procedure TPColor.ZoomAt(const X,Y : Double; const AHor,AVer : Double);
var
  Xt,Yt,Hor,Ver,XC,YC : Double;
begin
  Hor := Abs(AHor);
  Ver := Abs(AVer);
  XC := X;
  YC := Y;
  if Hor < 1 then Hor := 1;
  if Hor > MaxZoom then Hor := MaxZoom;
  if Ver < 1 then Ver := 1;
  if Ver > MaxZoom then Ver := MaxZoom;
  Xt := (FAxes.XMax - FAxes.XMin) / (2 * Hor);
  Yt := (FAxes.YMax - FAxes.YMin) / (2 * Ver);
  if (XC - Xt) < FAxes.XMin then XC := FAxes.XMin + Xt;
  if (XC + Xt) > FAxes.XMax then XC := FAxes.XMax - Xt;
  if (YC - Yt) < FAxes.YMin then YC := FAxes.YMin + Yt;
  if (YC + Yt) > FAxes.YMax then YC := FAxes.YMax - Yt;
  if (Hor <> FHorZoom) or (Ver <> FVerZoom)
      or ((XC - Xt) <> FLogLeft) or ((XC + Xt) <> FLogRight)
      or ((YC - Yt) <> FLogBottom) or ((YC + Yt) <> FLogTop) then
  begin
    FHorZoom := Hor;
    FVerZoom := Ver;
    FLogCentreX := XC;
    FLogCentreY := YC;
    FLogLeft := XC - Xt;
    FLogRight := XC + Xt;
    FLogTop := YC + Yt;
    FLogBottom := YC - Yt;
    CreateImage;
    RePaintImage;
    Zoom;
  end;  
end;

procedure TPColor.ZoomAtPos(const X,Y : Double);
begin
  ZoomAt(X,Y,HorZoom,VerZoom);
end;

procedure TPColor.SetLogCentreX(Value : Double);
begin
  ZoomAt(Value,LogCentreY,HorZoom,VerZoom);
end;

procedure TPColor.SetLogCentreY(Value : Double);
begin
  ZoomAt(LogCentreX,Value,HorZoom,VerZoom);
end;

procedure TPColor.ZoomAtTime(const AHor,AVer : Double);
begin
  ZoomAt(LogCentreX,LogCentreY,AHor,AVer);
end;

procedure TPColor.SetHorZoom(Value : Double);
begin
  ZoomAt(LogCentreX,LogCentreY,Value,VerZoom);
end;

procedure TPColor.SetVerZoom(Value : Double);
begin
  ZoomAt(LogCentreX,LogCentreY,HorZoom,Value);
end;
   
procedure TPColor.ZoomOut;
begin
  ZoomAtTime(1,1);
end;

procedure TPColor.ZoomArea(const ALeft,ATop,ARight,ABottom : Double);
var
  Hor,Ver,AL,AT,AR,AB : Double;
begin
  if ALeft < FAxes.XMin then AL := FAxes.XMin
   else if ALeft > FAxes.XMax then AL := FAxes.XMax
        else AL := ALeft;
  if ARight > FAxes.XMax then AR := FAxes.XMax
   else if ARight < FAxes.XMin then AR := FAxes.XMin
        else AR := ARight;
  if ATop > FAxes.YMax then AT := FAxes.YMax
   else if ATop < FAxes.YMin then AT := FAxes.YMin
        else AT := ATop;
  if ABottom < FAxes.YMin then AB := FAxes.YMin
   else if ABottom > FAxes.YMax then AB := FAxes.YMax
        else AB := ABottom;
  if AR = AL then Hor := MaxZoom
   else Hor := (FAxes.XMax - FAxes.XMin)/(AR - AL);
  if AT = AB then Ver := MaxZoom
   else Ver := (FAxes.YMax - FAxes.YMin)/(AT - AB);
  ZoomAt((ALeft+ARight)/2, (ATop+ABottom)/2, Hor, Ver);
end;

procedure TPColor.SetLogLeft(Value : Double);
begin
  ZoomArea(Value,LogTop,LogRight,LogBottom);
end;

procedure TPColor.SetLogTop(Value : Double);
begin
  ZoomArea(LogLeft,Value,LogRight,LogBottom);
end;

procedure TPColor.SetLogRight(Value : Double);
begin
  ZoomArea(LogLeft,LogTop,Value,LogBottom);
end;

procedure TPColor.SetLogBottom(Value : Double);
begin
  ZoomArea(LogLeft,LogTop,LogRight,Value);
end;

procedure TPColor.DrawCurColor(AColor : Integer);
var
  Step : double;
begin
  if ColorBarVis and (AColor <> 0) and (AColor <> 255) then
  with Canvas do
  begin
    if AColor = 253 then AColor := 0;
    if AColor = 254 then AColor := NumbCol + 1;
    if FColorBarKind = sbVertical then
    begin
      Step := FCoordinates.ColorBarHeight / (NumbCol + 2);
      Pen.Color := clBlack;
      Pen.Style := psSolid;
      Pen.Mode := pmNotXor;
      Brush.Style := bsClear;
      Rectangle(FCoordinates.ColorBarLeft - 1,
          FCoordinates.ColorBarTop + Trunc((NumbCol + 1 - AColor) * Step) - 1,
          FCoordinates.ColorBarLeft + FCoordinates.ColorBarWidth + 1,
          FCoordinates.ColorBarTop + Trunc((NumbCol + 2 - AColor) * Step) + 1);
    end
    else
    begin
      Step := FCoordinates.ColorBarWidth / (NumbCol + 2);
      Pen.Color := clBlack;
      Pen.Style := psSolid;
      Pen.Mode := pmNotXor;
      Brush.Style := bsClear;
      Rectangle(FCoordinates.ColorBarLeft + Trunc(AColor * Step) - 1,
          FCoordinates.ColorBarTop - 1,
          FCoordinates.ColorBarLeft + Trunc((AColor + 1) * Step) + 1,
          FCoordinates.ColorBarTop + FCoordinates.ColorBarHeight + 1);
    end;
  end;
end;

procedure TPColor.SetMouseOn(const Value: Boolean);
begin
  if FMouseOn <> Value then
  begin
    FMouseOn := Value;
    DrawCurColor(FOldColor);
  end;
end;

procedure TPColor.MouseDown(Button: TMouseButton; Shift: TShiftState;
   X, Y: Integer);
begin
  if (Button = mbLeft) and (NumberX <> 0) and (NumberY <> 0) and
     MouseOn and IsPosInImage(X, Y) then
  begin
    Drawing := True;
    Origin := Point(X, Y);
    MovePt := Origin;
  end;  
  if Assigned(FOnMouseDown) then FOnMouseDown(Self, Button, Shift, X, Y);
end;

procedure TPColor.MouseMove(Shift: TShiftState; X, Y: Integer);
begin
  if MouseOn and IsPosInImage(X, Y) then
  begin
    DrawCurColor(FOldColor);
    FOldColor := FDIB.Pixels[X-FCoordinates.ImageLeft,
        FCoordinates.ImageHeight-Y+FCoordinates.ImageTop-1];
    DrawCurColor(FOldColor);
    if Drawing then
    begin
      Canvas.Pen.Color := clBlack;
      Canvas.Pen.Style := psDot;
      Canvas.Pen.Mode := pmNotXor;
      Canvas.Brush.Style := bsClear;
      Canvas.Rectangle(Origin.X, Origin.Y, MovePt.X, MovePt.Y);
      MovePt := Point(X, Y);
      Canvas.Rectangle(Origin.X, Origin.Y, MovePt.X, MovePt.Y);
    end;
  end;
  if Assigned(FOnMouseMove) then FOnMouseMove(Self, Shift, X, Y);
end;


procedure TPColor.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  XA, YA, X1, Y1 : Double;
begin
  if (NumberX <> 0) and (NumberY <> 0) and MouseOn then
  begin
   if (Origin.X <> MovePt.X) or (Origin.Y <> MovePt.Y) then
   with Canvas do
   begin
     Pen.Color := clBlack;
     Pen.Style := psDot;
     Pen.Mode := pmNotXor;
     Brush.Style := bsClear;
     Rectangle(Origin.X, Origin.Y, MovePt.X, MovePt.Y);
   end;
   if ((Button = mbLeft) and (ssRight in Shift))
      or ((Button = mbRight) and (ssLeft in Shift)) then
   begin
     ZoomOut;
     MovePt := Origin;
   end
   else if ((Abs(Origin.X - MovePt.X) > MouseSensitivityX)
        or (Abs(Origin.Y - MovePt.Y) > MouseSensitivityY))
        and not((ssShift in Shift) or (ssAlt in Shift)
                  or (ssCtrl in Shift)) then
   begin
     PosImageToPosLog(Origin.x, Origin.y, X1, Y1);
     PosImageToPosLog(MovePt.x, MovePt.y, XA, YA);
     ZoomArea(X1, Y1, XA, YA);
     MovePt := Origin;
   end
   else if PosImageToPosLog(X, Y, XA, YA) then
     if (ssShift in Shift) and (Button = mbLeft) then ZoomAtPos(XA, YA)
     else if not((ssShift in Shift) or (ssAlt   in Shift)
                  or (ssCtrl in Shift)) then
          if (Button = mbLeft) then ZoomAt(XA, YA, HorZoom * 2, VerZoom * 2)
          else if (Button = mbRight) then ZoomAt(XA, YA, HorZoom/2, VerZoom/2);
  end;
  Drawing := False;
  if Assigned(FOnMouseUp) then FOnMouseUp(Self, Button, Shift, X, Y);
end;

procedure TPColor.SaveToFile(const FileName: String);
var
  TempBitmap: Graphics.TBitmap;
begin
  TempBitmap := TBitmap.Create;
  TempBitmap.Width := Width;
  TempBitmap.Height := Height;
  Draw(TempBitmap.Canvas);
  try
    TempBitmap.SaveToFile(FileName);
  finally
    TempBitmap.Free;
  end;
end;

procedure TPColor.SaveToClipBoard;
var
  TempBitmap: Graphics.TBitmap;
begin
  TempBitmap := TBitmap.Create;
  TempBitmap.Width := Width;
  TempBitmap.Height := Height;
  Draw(TempBitmap.Canvas);
  try
    ClipBoard.Assign(TempBitmap);
  finally
    TempBitmap.Free;
  end;
end;

procedure TPColor.SaveImageToFile(const FileName: String);
var
  TempBitmap: Graphics.TBitmap;
begin
  TempBitmap := TBitmap.Create;
  TempBitmap.Width := FCoordinates.ImageWidth;
  TempBitmap.Height := FCoordinates.ImageHeight;
  FDIB.Paint(TempBitmap.Canvas.Handle, 0, 0);
  try
    TempBitmap.SaveToFile(FileName);
  finally
    TempBitmap.Free;
  end;
end;

procedure TPColor.SaveImageToClipBoard;
var
  TempBitmap: Graphics.TBitmap;
begin
  TempBitmap := TBitmap.Create;
  TempBitmap.Width := FCoordinates.ImageWidth;
  TempBitmap.Height := FCoordinates.ImageHeight;
  FDIB.Paint(TempBitmap.Canvas.Handle, 0, 0);    
  try
    ClipBoard.Assign(TempBitmap);
  finally
    TempBitmap.Free;
  end;
end;

procedure TPColor.SetDrawColorBar(ADraw : Boolean);
begin
  if ADraw <> FDrawColorBar then
  begin
    FDrawColorBar := ADraw;
    Sizes;
    Paint;
  end;
end;

procedure TPColor.SetColorBarKind(AKind : TScrollBarKind);
begin
  if AKind <> FColorBarKind then
  begin
    FColorBarKind := AKind;
    Sizes;
    Paint;
  end;
end;

procedure TPColor.SetNumbCol(ANumbCol : Integer);
begin
  if ANumbCol < 2 then ANumbCol := 2;
  if ANumbCol > 252 then ANumbCol := 252;
  if ANumbCol <> FNumbCol then
  begin
    FNumbCol := ANumbCol;
    ChangeGraph := True;
    ChangePalette := True;
    Paint;
  end;
end;

function TPColor.GetUserPaletteEntry(Color: Byte): TPaletteEntry;
begin
  if Color > 251 then Color := 251;
  GetUserPaletteEntry := FUserPalette.palEntry[Color+1];
end;

procedure TPColor.SetUserPaletteEntry(Color: Byte; Value: TPaletteEntry);
begin
  if Color < 252 then
  begin
    FUserPalette.palEntry[Color + 1] := Value;
    FUserPalette.palEntry[Color + 1].peFlags := pc_Reserved;
    if ColorMap = cmUser then ChangePalette := True;
  end;
end;

procedure TPColor.CoordinatesChanged(Sender: TObject);
begin
  Sizes;
  Paint;
end;

procedure TPcolor.SetCoordinates(Value: TPColorCoordinates);
begin
  FCoordinates.Assign(Value);
end;

{ End TPColor }

end.
